import unittest

from hololinked.serializers import Serializers
from hololinked.serializers.serializers import BaseSerializer

try:
    from .utils import TestRunner, TestCase
    from .things import TestThing
except ImportError:
    from utils import TestRunner, TestCase
    from things import TestThing


class TestSerializer(TestCase):
    """Test the Serializers class"""

    # test register a new serializer with content type
    class YAMLSerializer(BaseSerializer):
        """just a dummy, does not really serialize to YAML"""

        @property
        def content_type(self):
            return "application/yaml"

    def test_1_singleton(self):
        """Test the singleton nature of the Serializers class."""

        serializers = Serializers()
        self.assertEqual(serializers, Serializers())
        self.assertNotEqual(Serializers, Serializers())
        self.assertIsInstance(serializers, Serializers)
        # all are class attributes
        self.assertEqual(serializers.json, Serializers.json)
        self.assertEqual(serializers.pickle, Serializers.pickle)
        self.assertEqual(serializers.msgpack, Serializers.msgpack)
        self.assertEqual(serializers.content_types, Serializers.content_types)
        self.assertEqual(serializers.object_content_type_map, Serializers.object_content_type_map)
        self.assertEqual(serializers.object_serializer_map, Serializers.object_serializer_map)
        self.assertEqual(serializers.protocol_serializer_map, Serializers.protocol_serializer_map)
        # check existing serializers are all instances of BaseSerializer
        for name, serializer in Serializers.content_types.items():
            self.assertIsInstance(serializer, BaseSerializer)
        # check default serializer, given that we know its JSON at least for the current test
        self.assertEqual(serializers.default, Serializers.json)
        self.assertEqual(serializers.default, Serializers.default)
        self.assertEqual(serializers.default, Serializers().json)
        self.assertEqual(serializers.default, Serializers().default)
        # check default content type, given that we know its JSON at least for the current test
        self.assertEqual(serializers.default_content_type, Serializers.json.content_type)
        # change default to pickle and check if it is set correctly
        # serializers.default = serializers.pickle
        # self.assertEqual(serializers.default, Serializers.pickle)
        # self.assertEqual(Serializers().default, Serializers.pickle)

    def test_2_protocol_registration(self):
        """i.e. test if a new serializer (protocol) can be registered"""

        # get existing number of serializers
        num_serializers = len(Serializers.content_types)

        # test register a new serializer
        base_serializer = BaseSerializer()
        # register with name
        self.assertWarns(UserWarning, Serializers.register, base_serializer, "base")
        # user warning because content type property is not defined
        # above is same as Serializers.register(base_serializer, 'base')

        # check if name became a class attribute and name can be accessed as an attribute
        self.assertIn("base", Serializers)
        self.assertEqual(Serializers.base, base_serializer)
        self.assertEqual(Serializers().base, base_serializer)
        # we dont support getitem at instance level yet so we cannot test assertIn

        # since a content type is not set, it should not be in the content types
        self.assertNotIn(base_serializer, Serializers.content_types.values())
        # so the length of content types should be the same
        self.assertEqual(len(Serializers.content_types), num_serializers)

        # instantiate
        yaml_serializer = self.YAMLSerializer()
        # register with name
        Serializers.register(yaml_serializer, "yaml")
        # check if name became a class attribute and name can be accessed as an attribute
        self.assertIn("yaml", Serializers)
        self.assertEqual(Serializers.yaml, yaml_serializer)
        self.assertEqual(Serializers().yaml, yaml_serializer)
        # we dont support getitem at instance level yet

        # since a content type is set, it should be in the content types
        self.assertIn(yaml_serializer.content_type, Serializers.content_types.keys())
        self.assertIn(yaml_serializer, Serializers.content_types.values())
        # so the length of content types should have increased by 1
        self.assertEqual(len(Serializers.content_types), num_serializers + 1)

    def test_3_registration_for_objects(self):
        """i.e. test if a new serializer can be registered for a specific property, action or event"""
        Serializers.register_content_type_for_object(TestThing.base_property, "application/x-pickle")
        Serializers.register_content_type_for_object(TestThing.action_echo, "application/msgpack")
        Serializers.register_content_type_for_object(TestThing.test_event, "application/yaml")

        self.assertEqual(
            Serializers.for_object(None, "TestThing", "action_echo"),
            Serializers.msgpack,
        )
        self.assertEqual(
            Serializers.for_object(None, "TestThing", "base_property"),
            Serializers.pickle,
        )
        self.assertEqual(Serializers.for_object(None, "TestThing", "test_event"), Serializers.yaml)
        self.assertEqual(
            Serializers.for_object(None, "TestThing", "test_unknown_property"),
            Serializers.default,
        )

    def test_4_registration_for_objects_by_name(self):
        Serializers.register_content_type_for_object_per_thing_instance(
            "test_thing", "base_property", "application/yaml"
        )
        self.assertIsInstance(
            Serializers.for_object("test_thing", None, "base_property"),
            self.YAMLSerializer,
        )

    def test_5_registration_dict(self):
        """test the dictionary where all serializers are stored"""
        # depends on test 3
        self.assertIn("test_thing", Serializers.object_content_type_map)
        self.assertIn("base_property", Serializers.object_content_type_map["test_thing"])
        self.assertEqual(
            Serializers.object_content_type_map["test_thing"]["base_property"],
            "application/yaml",
        )

        self.assertIn("action_echo", Serializers.object_content_type_map["TestThing"])
        self.assertEqual(
            Serializers.object_content_type_map["TestThing"]["action_echo"],
            "application/msgpack",
        )
        self.assertIn("test_event", Serializers.object_content_type_map["TestThing"])
        self.assertEqual(
            Serializers.object_content_type_map["TestThing"]["test_event"],
            "application/yaml",
        )

    def test_6_retrieval(self):
        # added in previous tests
        self.assertIsInstance(
            Serializers.for_object("test_thing", None, "base_property"),
            self.YAMLSerializer,
        )
        # unknown object should retrieve the default serializer
        self.assertEqual(
            Serializers.for_object("test_thing", None, "test_unknown_property"),
            Serializers.default,
        )
        # unknown thing should retrieve the default serializer
        self.assertEqual(
            Serializers.for_object("test_unknown_thing", None, "base_property"),
            Serializers.default,
        )

    def test_7_set_default(self):
        """test setting the default serializer"""
        # get existing default
        old_default = Serializers.default
        # set new default and check if default is set
        Serializers.default = Serializers.yaml
        self.assertEqual(Serializers.default, Serializers.yaml)
        self.test_6_retrieval()  # check if retrieval is consistent with default
        # reset default and check if default is reset
        Serializers.default = old_default
        self.assertEqual(Serializers.default, old_default)
        self.assertEqual(Serializers.default, Serializers.json)  # because we know its JSON

    @classmethod
    def tearDownClass(cls):
        Serializers.reset()
        return super().tearDownClass()


if __name__ == "__main__":
    unittest.main(testRunner=TestRunner())



from .brokers import (  # noqa: F401
    AsyncZMQServer,
    ZMQServerPool,
    SyncZMQClient,
    AsyncZMQClient,
    MessageMappedZMQClientPool,
    EventPublisher,
    AsyncEventConsumer,
    EventConsumer,
)



import typing
from pydantic import Field

from .base import Schema


class Link(Schema):
    href: str
    anchor: typing.Optional[str]
    rel: typing.Optional[str]
    type: typing.Optional[str] = Field(default="application/json")


class VersionInfo(Schema):
    """
    create version info.
    schema - https://www.w3.org/TR/wot-thing-description11/#versioninfo
    """

    instance: str
    model: str



import copy
import socket
import zmq
import zmq.asyncio
import asyncio
import typing
import threading
import logging
import tracemalloc
from collections import deque


from ...exceptions import BreakLoop, BreakInnerLoop
from ...constants import ZMQ_TRANSPORTS, Operations
from ...utils import (
    format_exception_as_json,
    get_all_sub_things_recusively,
    get_current_async_loop,
    get_default_logger,
    set_global_event_loop_policy,
)
from ...config import global_config
from ...serializers import Serializers, BaseSerializer
from .message import (
    EMPTY_BYTE,
    ERROR,
    REPLY,
    PreserializedData,
    RequestMessage,
    SerializableData,
)
from .brokers import AsyncZMQServer, BaseZMQServer, EventPublisher
from ..thing import Thing
from ..property import Property  # noqa: F401
from ..properties import TypedDict
from ..actions import BoundAction  # noqa: F401
from ..logger import LogHistoryHandler


if global_config.TRACE_MALLOC:
    tracemalloc.start()

set_global_event_loop_policy()

Undefined = NotImplemented


class RPCServer(BaseZMQServer):
    """
    The `RPCServer` implements a infinite loop where ZMQ sockets listen for messages, in any transport layer possible
    (`INPROC`, `IPC` or `TCP`). Once requests are received, jobs are dispatched to the `Thing` instances which are being served,
    with timeouts or any other execution requirements (called execution context). After being executed by a `Thing` instance,
    the results are then sent back to the client. Execution information include `Thing` ID, the property, action or
    event to be executed (events are usually PUB-SUB and are largely handled by the `EventPublisher` directly),
    what operation to do on them (i.e. `readproperty`, `invokeaction` etc.), the payload and the execution contexts (like timeouts).
    This is structured as a JSON.

    Jobs determine how to execute the operations on the `Thing` instance, whether in queued, async or threaded modes.
    This is their main function. Queued mode is the default as it is assumed that multiple physical operations in the physical world
    is not always practical.

    Default ZMQ transport layer is `INPROC`, but `IPC` or `TCP` can also be added simultaneously using `ZMQServer`
    instance instead of `RPCServer`. The purpose of `INPROC` being default is that, the `RPCServer` is the only server
    implementing the operations directly on the `Thing`
    instances. All other protocols like HTTP, MQTT, CoAP etc. will be used to redirect requests to the `RPCServer` only
    and do not directly operate on the `Thing` instances. Instead, the incoming requests in those protocols are converted
    to the above stated "Jobs" which are in JSON format.

    `INPROC` is the fastest and most efficient way to communicate between multiple independently running loops,
    whether the loop belongs to a specific protocol's request listener or the `RPCServer` itself, as it used shared memory.
    The same `INPROC` messaging contract is also used for `IPC` and `TCP`, thus eliminating the
    need to separately implement messaging contracts at different layers of communication for ZMQ.

    Therefore, if a `Thing` instance is to be served by a well known protocol, say HTTP, the server behaves like HTTP-RPC.

    [UML Diagram](http://docs.hololinked.dev/UML/PDF/RPCServer.pdf)
    """

    things = TypedDict(
        key_type=(str,),
        item_type=(Thing,),
        bounds=(0, 100),
        allow_None=True,
        default=None,
        doc="list of Things which are being executed",
        remote=False,
    )  # type: typing.Dict[str, Thing]

    def __init__(
        self,
        *,
        id: str,
        things: typing.List[Thing],
        context: zmq.asyncio.Context | None = None,
        access_point: ZMQ_TRANSPORTS = ZMQ_TRANSPORTS.INPROC,
        **kwargs: typing.Dict[str, typing.Any],
    ) -> None:
        """
        Parameters
        ----------
        id: str
            `id` of the server
        things: List[Thing]
            list of `Thing` instances to be served
        context: Optional, zmq.asyncio.Context
            ZeroMQ async Context object to use. Automatically created when None is supplied.
        access_point: ZMQ_TRANSPORTS
            transport layer to be used for the server, default is `INPROC`
        """
        super().__init__(id=id, **kwargs)
        self.things = dict()
        for thing in things:
            self.things[thing.id] = thing

        if isinstance(access_point, str):
            access_point = access_point.upper()

        if self.logger is None:
            self.logger = get_default_logger(
                "{}|{}|{}|{}".format(self.__class__.__name__, "RPC", "MIXED", self.id),
                kwargs.get("log_level", logging.INFO),
            )
            kwargs["logger"] = self.logger

        self._run = False  # flag to stop all the servers
        self.context = context or global_config.zmq_context()

        self.req_rep_server = AsyncZMQServer(
            id=self.id,
            context=self.context,
            access_point=access_point,
            poll_timeout=1000,
            **kwargs,
        )
        self.event_publisher = EventPublisher(
            id=f"{self.id}/event-publisher",
            context=self.context,
            access_point=access_point,
            **kwargs,
        )
        self.schedulers = dict()

        # setup scheduling requirements
        for instance in self.things.values():
            all_things = get_all_sub_things_recusively(instance)
            for instance in all_things:
                assert isinstance(instance, Thing), "instance must be of type Thing"
                instance.rpc_server = self
                for action in instance.actions.descriptors.values():
                    if action.execution_info.iscoroutine and not action.execution_info.synchronous:
                        self.schedulers[f"{instance.id}.{action.name}.invokeAction"] = AsyncScheduler
                    elif not action.execution_info.synchronous:
                        self.schedulers[f"{instance.id}.{action.name}.invokeAction"] = ThreadedScheduler
                    # else QueuedScheduler which is default
                # properties need not dealt yet, but may be in future

    schedulers: typing.Dict[str, "QueuedScheduler"]

    def __post_init__(self):
        super().__post_init__()
        self.logger.info("Server with name '{}' can be started using run().".format(self.id))

    @property
    def is_running(self) -> bool:
        """Check if the server is running or not."""
        return self._run

    async def recv_requests_and_dispatch_jobs(self, server: AsyncZMQServer) -> None:
        """
        Continuously receives messages from different clients and dispatches them as jobs according to the specific
        requirements of a how an object (property/action/event) must be executed (queued/threaded/async).
        Also handles messages that dont need separate jobs like `HANDSHAKE`, `EXIT`, timeouts etc.

        Parameters
        ----------
        server: AsyncZMQServer
            the server instance to poll for requests
        """
        self.logger.debug("started polling with server {} at socket {}".format(server.id, server.socket_address))
        eventloop = asyncio.get_event_loop()
        while self._run:
            try:
                request_messages = await server.poll_requests()
                # when stop poll is set, this will exit with an empty list
            except BreakLoop:
                break
            except Exception as ex:
                self.logger.error(f"exception occurred while polling for server '{server.id}' - {str(ex)}")
                continue

            for request_message in request_messages:
                try:
                    # handle invokation timeout
                    invokation_timeout = request_message.server_execution_context.get("invokation_timeout", None)

                    ready_to_process_event = None
                    timeout_task = None
                    if invokation_timeout is not None:
                        ready_to_process_event = asyncio.Event()
                        timeout_task = asyncio.create_task(
                            self._process_timeouts(
                                request_message=request_message,
                                ready_to_process_event=ready_to_process_event,
                                timeout=invokation_timeout,
                                origin_server=server,
                                timeout_type="invokation",
                            )
                        )
                        eventloop.call_soon(lambda: timeout_task)

                    # check object level scheduling requirements and schedule the message
                    # append to messages list - message, event, timeout task, origin socket
                    job = (
                        server,
                        request_message,
                        timeout_task,
                        ready_to_process_event,
                    )  # type: Scheduler.JobInvokationType
                    if request_message.qualified_operation in self.schedulers:
                        scheduler = self.schedulers[request_message.qualified_operation](
                            self.things[request_message.thing_id], self
                        )
                    else:
                        scheduler = self.schedulers[request_message.thing_id]
                    scheduler.dispatch_job(job)

                except Exception as ex:
                    # handle invalid message
                    self.logger.error(f"exception occurred for message id '{request_message.id}' - {str(ex)}")
                    invalid_message_task = asyncio.create_task(
                        server._handle_invalid_message(request_message=request_message, exception=ex)
                    )
                    eventloop.call_soon(lambda: invalid_message_task)
        self.stop()
        self.logger.info(f"stopped polling for server '{server.id}' {server.socket_address.split(':')[0].upper()}")

    async def tunnel_message_to_things(self, scheduler: "Scheduler") -> None:
        """message tunneler/coordinator between external sockets listening thread and `Thing` object executor thread"""
        eventloop = get_current_async_loop()
        while self._run and scheduler.run:
            # wait for message first
            if not scheduler.has_job:
                await scheduler.wait_for_job()
                # this means in next loop it wont be in this block as a message arrived
                continue

            # retrieve from messages list - message, execution context, event, timeout task, origin socket
            origin_server, request_message, timeout_task, ready_to_process_event = scheduler.next_job
            server_execution_context = request_message.server_execution_context

            # handle invokation timeout
            invokation_timed_out = True
            if ready_to_process_event is not None:
                ready_to_process_event.set()  # releases timeout task
                invokation_timed_out = await timeout_task
            if ready_to_process_event is not None and invokation_timed_out:
                # drop call to thing, timeout message was already sent in _process_timeouts()
                continue

            # handle execution through thing
            scheduler.last_operation_request = scheduler.extract_operation_tuple_from_request(request_message)

            # schedule an execution timeout
            execution_timeout = server_execution_context.get("execution_timeout", None)
            execution_completed_event = None
            execution_timeout_task = None
            execution_timed_out = True
            if execution_timeout is not None:
                execution_completed_event = asyncio.Event()
                execution_timeout_task = asyncio.create_task(
                    self._process_timeouts(
                        request_message=request_message,
                        ready_to_process_event=execution_completed_event,
                        timeout=execution_timeout,
                        origin_server=origin_server,
                        timeout_type="execution",
                    )
                )
                eventloop.call_soon(lambda: execution_timeout_task)

            # always wait for reply from thing, since this loop is asyncio task (& in its own thread in RPC server),
            # timeouts always reach client without truly blocking by the GIL. If reply does not arrive, all other requests
            # get invokation timeout.
            # await eventloop.run_in_executor(None, scheduler.wait_for_reply)
            await scheduler.wait_for_reply(eventloop)
            # check if reply is never undefined, Undefined is a sensible placeholder for NotImplemented singleton
            if scheduler.last_operation_reply is Undefined:
                # this is a logic error, as the reply should never be undefined
                await origin_server._handle_error_message(
                    request_message=request_message,
                    exception=RuntimeError("No reply from thing - logic error"),
                )
                continue
            payload, preserialized_payload, reply_message_type = scheduler.last_operation_reply
            scheduler.reset_operation_reply()

            # check if execution completed within time
            if execution_completed_event is not None:
                execution_completed_event.set()  # releases timeout task
                execution_timed_out = await execution_timeout_task
            if execution_timeout_task is not None and execution_timed_out:
                # drop reply to client as timeout was already sent
                continue
            if server_execution_context.get("oneway", False):
                # drop reply if oneway
                continue

            # send reply to client
            await origin_server.async_send_response_with_message_type(
                request_message=request_message,
                message_type=reply_message_type,
                payload=payload,
                preserialized_payload=preserialized_payload,
            )

        scheduler.cleanup()
        self.logger.info("stopped tunneling messages to things")

    async def run_thing_instance(self, instance: Thing, scheduler: typing.Optional["Scheduler"] = None) -> None:
        """
        run a single `Thing` instance in an infinite loop by allowing the scheduler to schedule operations on it.

        Parameters
        ----------
        instance: Thing
            instance of the `Thing`
        scheduler: Optional[Scheduler]
            scheduler that schedules operations on the `Thing` instance, a default is always available.
        """
        self.logger.info(
            "starting to run operations on thing {} of class {}".format(instance.id, instance.__class__.__name__)
        )
        if self.logger.level >= logging.ERROR:
            # sleep added to resolve some issue with logging related IO bound tasks in asyncio - not really clear what it is.
            # This loop crashes for log levels above ERROR without the following statement
            await asyncio.sleep(0.001)
            # TODO - investigate and fix it
        scheduler = scheduler or self.schedulers[instance.id]
        eventloop = get_current_async_loop()

        while self._run and scheduler.run:
            # print("starting to serve thing {}".format(instance.id))
            await scheduler.wait_for_operation(eventloop)
            # await scheduler.wait_for_operation()
            if scheduler.last_operation_request is Undefined:
                instance.logger.warning("No operation request found in thing '{}'".format(instance.id))
                continue

            try:
                # fetch operation_request which is a tuple of
                # (thing_id, objekt, operation, payload, preserialized_payload, execution_context)
                (
                    thing_id,
                    objekt,
                    operation,
                    payload,
                    preserialized_payload,
                    execution_context,
                ) = scheduler.last_operation_request

                # deserializing the payload required to execute the operation
                payload = payload.deserialize()
                preserialized_payload = preserialized_payload.value
                instance.logger.debug(
                    f"thing {instance.id} with {thing_id} starting execution of operation {operation} on {objekt}"
                )

                # start activities related to thing execution context
                fetch_execution_logs = execution_context.pop("fetch_execution_logs", False)
                if fetch_execution_logs:
                    list_handler = LogHistoryHandler([])
                    list_handler.setLevel(logging.DEBUG)
                    list_handler.setFormatter(instance.logger.handlers[0].formatter)
                    instance.logger.addHandler(list_handler)

                # execute the operation
                return_value = await self.execute_operation(instance, objekt, operation, payload, preserialized_payload)

                # handle return value
                serializer = Serializers.for_object(thing_id, instance.__class__.__name__, objekt)
                rpayload, rpreserialized_payload = self.format_return_value(return_value, serializer=serializer)

                # complete thing execution context
                if fetch_execution_logs:
                    rpayload.value = dict(return_value=rpayload.value, execution_logs=list_handler.log_list)

                # raise any payload errors now
                rpayload.require_serialized()

                # set reply
                scheduler.last_operation_reply = (rpayload, rpreserialized_payload, REPLY)

            except BreakInnerLoop:
                # exit the loop and stop the thing
                instance.logger.info(
                    "Thing {} with id {} exiting event loop.".format(instance.__class__.__name__, instance.id)
                )

                # send a reply with None return value
                rpayload, rpreserialized_payload = self.format_return_value(None, Serializers.json)

                # complete thing execution context
                if fetch_execution_logs:
                    rpayload.value = dict(return_value=rpayload.value, execution_logs=list_handler.log_list)

                # set reply, let the message broker decide
                scheduler.last_operation_reply = (rpayload, rpreserialized_payload, None)

                # quit the loop
                break

            except Exception as ex:
                # error occurred while executing the operation
                instance.logger.error(
                    "Thing {} with ID {} produced error : {} - {}.".format(
                        instance.__class__.__name__, instance.id, type(ex), ex
                    )
                )

                # send a reply with error
                rpayload, rpreserialized_payload = self.format_return_value(
                    dict(exception=format_exception_as_json(ex)), Serializers.json
                )

                # complete thing execution context
                if fetch_execution_logs:
                    rpayload.value["execution_logs"] = list_handler.log_list

                # set error reply
                scheduler.last_operation_reply = (rpayload, rpreserialized_payload, ERROR)

            finally:
                # cleanup
                if fetch_execution_logs:
                    instance.logger.removeHandler(list_handler)
                instance.logger.debug(
                    "thing {} with id {} completed execution of operation {} on {}".format(
                        instance.__class__.__name__, instance.id, operation, objekt
                    )
                )
        self.logger.info("stopped running thing {}".format(instance.id))

    async def execute_operation(
        self,
        instance: Thing,
        objekt: str,
        operation: str,
        payload: typing.Any,
        preserialized_payload: bytes,
    ) -> typing.Any:
        """
        Execute a given operation on a thing instance.

        Parameters
        ----------
        instance: Thing
            instance of the thing
        objekt: str
            name of the property, action or event
        operation: str
            operation to be executed on the property, action or event
        payload: Any
            payload to be used for the operation
        preserialized_payload: bytes
            preserialized payload to be used for the operation
        """
        if operation == Operations.readproperty:
            prop = instance.properties[objekt]  # type: Property
            return getattr(instance, prop.name)
        elif operation == Operations.writeproperty:
            prop = instance.properties[objekt]  # type: Property
            if preserialized_payload != EMPTY_BYTE:
                prop_value = preserialized_payload
            else:
                prop_value = payload
            return prop.external_set(instance, prop_value)
        elif operation == Operations.deleteproperty:
            prop = instance.properties[objekt]  # type: Property
            del prop  # raises NotImplementedError when deletion is not implemented which is mostly the case
        elif operation == Operations.invokeaction and objekt == "get_thing_description":
            # special case
            if payload is None:
                payload = dict()
            args = payload.pop("__args__", tuple())
            return self.get_thing_description(instance, *args, **payload)
        elif operation == Operations.invokeaction:
            if payload is None:
                payload = dict()
            args = payload.pop("__args__", tuple())
            # payload then become kwargs
            if preserialized_payload != EMPTY_BYTE:
                args = (preserialized_payload,) + args
            action = instance.actions[objekt]  # type: BoundAction
            if action.execution_info.iscoroutine:
                # the actual scheduling as a purely async task is done by the scheduler, not here,
                # this will be a blocking call
                return await action.external_call(*args, **payload)
            return action.external_call(*args, **payload)
        elif operation == Operations.readmultipleproperties or operation == Operations.readallproperties:
            if objekt is None:
                return instance.properties.get()
            return instance.properties.get(names=objekt)
        elif operation == Operations.writemultipleproperties or operation == Operations.writeallproperties:
            return instance.properties.set(payload)
        raise NotImplementedError(
            "Unimplemented execution path for Thing {} for operation {}".format(instance.id, operation)
        )

    def format_return_value(
        self,
        return_value: typing.Any,
        serializer: BaseSerializer,
    ) -> tuple[SerializableData, PreserializedData]:
        if (
            isinstance(return_value, tuple)
            and len(return_value) == 2
            and (isinstance(return_value[1], bytes) or isinstance(return_value[1], PreserializedData))
        ):
            payload = SerializableData(return_value[0], serializer=serializer, content_type=serializer.content_type)
            if isinstance(return_value[1], bytes):
                preserialized_payload = PreserializedData(return_value[1])
        elif isinstance(return_value, bytes):
            payload = SerializableData(None, content_type="application/json")
            preserialized_payload = PreserializedData(return_value)
        elif isinstance(return_value, PreserializedData):
            payload = SerializableData(None, content_type="application/json")
            preserialized_payload = return_value
        else:
            payload = SerializableData(return_value, serializer=serializer, content_type=serializer.content_type)
            preserialized_payload = PreserializedData(EMPTY_BYTE, content_type="text/plain")
        return payload, preserialized_payload

    async def _process_timeouts(
        self,
        request_message: RequestMessage,
        ready_to_process_event: asyncio.Event,
        origin_server: AsyncZMQServer,
        timeout: float | int | None,
        timeout_type: str,
    ) -> bool:
        """
        replies timeout to client if timeout occured along with returning `True` to indicate that.
        If timeout did not occur, the `ready_to_process_event` is set to indicate that the operation can be processed.
        `False` is returned in this case.
        """
        try:
            await asyncio.wait_for(ready_to_process_event.wait(), timeout)
            return False
        except TimeoutError:
            await origin_server._handle_timeout(request_message, timeout_type)
            return True

    def run_zmq_request_listener(self):
        """
        Runs ZMQ's socket polling in an async loop. This method is blocking and is automatically called by `run()`
        method. Please dont call this method when the async loop is already running.
        """
        self.logger.info("starting external message listener thread")
        eventloop = get_current_async_loop()
        existing_tasks = asyncio.all_tasks(eventloop)
        eventloop.run_until_complete(
            asyncio.gather(
                self.recv_requests_and_dispatch_jobs(self.req_rep_server),
                *[self.tunnel_message_to_things(scheduler) for scheduler in self.schedulers.values()],
                *existing_tasks,
            )
        )
        self.logger.info("exiting external listener event loop {}".format(self.id))
        eventloop.close()

    def run_things(self, things: typing.List[Thing]):
        """
        Run loop that executes operations on `Thing` instances. This method is blocking and is called by `run()` method.

        Parameters
        ----------
        things: List[Thing]
            list of `Thing` instances to be executed
        """
        thing_executor_loop = get_current_async_loop()
        self.logger.info(
            f"starting thing executor loop in thread {threading.get_ident()} for {[obj.id for obj in things]}"
        )
        thing_executor_loop.run_until_complete(
            asyncio.gather(*[self.run_thing_instance(instance) for instance in things])
        )
        self.logger.info(f"exiting event loop in thread {threading.get_ident()}")
        thing_executor_loop.close()

    def run(self):
        """
        Start & run the server. This method is blocking.
        Creates job schedulers for each `Thing`, dispatches each `Thing` to its own thread and starts the ZMQ sockets
        request polling loop. Call `stop()` (threadsafe) to stop the server.
        """
        self._run = True
        self.logger.info(f"starting RPC server {self.id}")
        for thing in self.things.values():
            self.schedulers[thing.id] = QueuedScheduler(thing, self)
        threads = dict()  # type: typing.Dict[int, threading.Thread]
        for thing in self.things.values():
            thread = threading.Thread(target=self.run_things, args=([thing],))
            thread.start()
            threads[thread.ident] = thread
        self.run_zmq_request_listener()
        for thread in threads.values():
            thread.join()
        self.logger.info(f"server stopped {self.id}")

    def stop(self):
        """Stop the server. This method is threadsafe."""
        self._run = False
        self.req_rep_server.stop_polling()
        for scheduler in self.schedulers.values():
            scheduler.cleanup()

    def exit(self):
        """Stop and try to clean up resources used by the server."""
        try:
            self.stop()
            if self.req_rep_server is not None:
                self.req_rep_server.exit()
            if self.event_publisher is not None:
                self.event_publisher.exit()
        except Exception as ex:
            self.logger.warning(f"Exception occurred while exiting the server - {str(ex)}")

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, RPCServer):
            return False
        return self.id == other.id

    def __str__(self):
        return f"RPCServer({self.id}, req-rep: {self.req_rep_server.socket_address}, pub-sub: {self.event_publisher.socket_address})"

    def get_thing_description(
        self,
        instance: Thing,
        protocol: str,
        ignore_errors: bool = False,
        skip_names: list[str] = [],
    ) -> dict[str, typing.Any]:
        """
        Get the Thing Description (TD) for a specific Thing instance.

        Parameters
        ----------
        instance: Thing
            The Thing instance for which to retrieve the TD
        protocol: str
            The protocol for which to generate the TD - `INPROC`, `IPC` or `TCP`
        ignore_errors: bool
            Whether to ignore errors while generating the TD. Default is False.
        skip_names: List[str]
            List of property, action or event names to skip while generating the TD. Default is empty list.

        Returns
        -------
        JSON
            The Thing Description in JSON format.
        """
        TM = instance.get_thing_model(ignore_errors=ignore_errors, skip_names=skip_names).json()  # type: dict[str, typing.Any]
        TD = copy.deepcopy(TM)
        from ...td import PropertyAffordance, ActionAffordance, EventAffordance
        from ...td.forms import Form

        if protocol.lower() == "inproc":
            req_rep_socket_address = self.req_rep_server.socket_address
            pub_sub_socket_address = self.event_publisher.socket_address
        elif protocol.lower() == "ipc":
            if not hasattr(self, "ipc_server"):
                raise RuntimeError(
                    "This server cannot generate TD for IPC protocol, consider using thing model directly."
                )
            req_rep_socket_address = self.ipc_server.socket_address
            pub_sub_socket_address = self.ipc_event_publisher.socket_address
        elif protocol.lower() == "tcp":
            if not hasattr(self, "tcp_server"):
                raise RuntimeError(
                    "This server cannot generate TD for TCP protocol, consider using thing model directly."
                )
            req_rep_socket_address = self.tcp_server.socket_address  # type: str
            req_rep_socket_address = req_rep_socket_address.replace("*", socket.gethostname()).replace(
                "0.0.0.0", socket.gethostname()
            )
            pub_sub_socket_address = self.tcp_event_publisher.socket_address  # type: str
            pub_sub_socket_address = pub_sub_socket_address.replace("*", socket.gethostname()).replace(
                "0.0.0.0", socket.gethostname()
            )
        else:
            raise ValueError(f"Unsupported protocol '{protocol}' for ZMQ.")

        for name in TM.get("properties", []):
            affordance = PropertyAffordance.from_TD(name, TM)
            if not TD["properties"][name].get("forms", None):
                TD["properties"][name]["forms"] = []

            form = Form()
            form.href = req_rep_socket_address
            form.op = Operations.readproperty
            form.contentType = Serializers.for_object(instance.id, instance.__class__.__name__, name).content_type
            TD["properties"][name]["forms"].append(form.json())

            if not affordance.readOnly:
                form = Form()
                form.href = req_rep_socket_address
                form.op = Operations.writeproperty
                form.contentType = Serializers.for_object(instance.id, instance.__class__.__name__, name).content_type
                TD["properties"][name]["forms"].append(form.json())

            if affordance.observable:
                form = Form()
                form.href = pub_sub_socket_address
                form.op = Operations.observeproperty
                form.contentType = Serializers.for_object(instance.id, instance.__class__.__name__, name).content_type
                TD["properties"][name]["forms"].append(form.json())

        for name in TM.get("actions", []):
            affordance = ActionAffordance.from_TD(name, TM)
            if not TD["actions"][name].get("forms", None):
                TD["actions"][name]["forms"] = []

            form = Form()
            form.href = req_rep_socket_address
            form.op = Operations.invokeaction
            form.contentType = Serializers.for_object(instance.id, instance.__class__.__name__, name).content_type
            TD["actions"][name]["forms"].append(form.json())

        for name in TM.get("events", []):
            affordance = EventAffordance.from_TD(name, TM)
            if not TD["events"][name].get("forms", None):
                TD["events"][name]["forms"] = []

            form = Form()
            form.href = pub_sub_socket_address
            form.op = Operations.subscribeevent
            form.contentType = Serializers.for_object(instance.id, instance.__class__.__name__, name).content_type
            TD["events"][name]["forms"].append(form.json())

        return TD


class Scheduler:
    """
    Scheduler class to schedule the operations of a thing either in queued mode, or a one-shot mode in either
    async or threaded loops.

    [UML Diagram](http://docs.hololinked.dev/UML/PDF/RPCServer.pdf)

    [UML Diagram subclasses](http://docs.hololinked.dev/UML/PDF/Scheduler.pdf)
    """

    OperationRequest = typing.Tuple[str, str, str, SerializableData, PreserializedData, typing.Dict[str, typing.Any]]
    OperationReply = typing.Tuple[SerializableData, PreserializedData, str]
    JobInvokationType = typing.Tuple[AsyncZMQServer, RequestMessage, asyncio.Task, asyncio.Event]
    # [UML Diagram](http://docs.hololinked.dev/UML/PDF/RPCServer.pdf)
    _operation_execution_complete_event: asyncio.Event | threading.Event
    _operation_execution_ready_event: asyncio.Event | threading.Event

    def __init__(self, instance: Thing, rpc_server: RPCServer) -> None:
        self.instance = instance  # type: Thing
        self.rpc_server = rpc_server  # type: RPCServer
        self.run = True  # type: bool
        self._one_shot = False  # type: bool
        self._last_operation_request = Undefined  # type: Scheduler.OperationRequest
        self._last_operation_reply = Undefined  # type: Scheduler.OperationRequest
        self._job_queued_event = asyncio.Event()  # type: asyncio.Event

    @property
    def last_operation_request(self) -> OperationRequest:
        return self._last_operation_request

    @last_operation_request.setter
    def last_operation_request(self, value: OperationRequest):
        self._last_operation_request = value
        self._operation_execution_ready_event.set()

    def reset_operation_request(self) -> None:
        self._last_operation_request = Undefined

    @property
    def last_operation_reply(self) -> OperationReply:
        return self._last_operation_reply

    @last_operation_reply.setter
    def last_operation_reply(self, value: OperationReply):
        self._last_operation_request = Undefined
        self._last_operation_reply = value
        self._operation_execution_complete_event.set()
        if self._one_shot:
            self.run = False

    def reset_operation_reply(self) -> None:
        self._last_operation_reply = Undefined

    async def wait_for_job(self) -> None:
        await self._job_queued_event.wait()
        self._job_queued_event.clear()

    async def wait_for_operation(self, eventloop: asyncio.AbstractEventLoop | None) -> None:
        # assert isinstance(self._operation_execution_ready_event, threading.Event), "not a threaded scheduler"
        if isinstance(self._operation_execution_ready_event, threading.Event):
            await eventloop.run_in_executor(None, self._operation_execution_ready_event.wait)
        else:
            await self._operation_execution_ready_event.wait()
        self._operation_execution_ready_event.clear()

    async def wait_for_reply(self, eventloop: asyncio.AbstractEventLoop | None) -> None:
        if isinstance(self._operation_execution_complete_event, threading.Event):
            await eventloop.run_in_executor(None, self._operation_execution_complete_event.wait)
        else:
            await self._operation_execution_complete_event.wait()
        self._operation_execution_complete_event.clear()

    @property
    def has_job(self) -> bool:
        raise NotImplementedError("has_job method must be implemented in the subclass")

    @property
    def next_job(self) -> JobInvokationType:
        raise NotImplementedError("next_job method must be implemented in the subclass")

    def dispatch_job(self, job: JobInvokationType) -> None:
        raise NotImplementedError("dispatch_job method must be implemented in the subclass")

    def cleanup(self):
        self.run = False
        self._job_queued_event.set()
        self._operation_execution_ready_event.set()
        self._operation_execution_complete_event.set()

    @classmethod
    def extract_operation_tuple_from_request(self, request_message: RequestMessage) -> OperationRequest:
        """thing execution info"""
        return (
            request_message.header["thingID"],
            request_message.header["objekt"],
            request_message.header["operation"],
            request_message.body[0],
            request_message.body[1],
            request_message.header["thingExecutionContext"],
        )

    @classmethod
    def format_reply_tuple(self, return_value: typing.Any) -> OperationReply:
        pass


class QueuedScheduler(Scheduler):
    """
    Scheduler class to schedule the operations of a thing in a queued loop.
    """

    def __init__(self, instance: Thing, rpc_server: RPCServer) -> None:
        super().__init__(instance, rpc_server)
        self.queue = deque()
        self._one_shot = False
        self._operation_execution_ready_event = threading.Event()
        self._operation_execution_complete_event = threading.Event()

    @property
    def has_job(self) -> bool:
        return len(self.queue) > 0

    @property
    def next_job(self) -> Scheduler.JobInvokationType:
        return self.queue.popleft()

    def dispatch_job(self, job: Scheduler.JobInvokationType) -> None:
        """
        append a request message to the queue after ticking the invokation timeout clock

        Parameters
        ----------
        job: Tuple[RequestMessage, asyncio.Event, asyncio.Task, AsyncZMQServer]
            tuple of request message, event to indicate if request message can be executed, invokation timeout task
            and originating server of the request
        """
        self.queue.append(job)
        self._job_queued_event.set()

    def cleanup(self):
        self.queue.clear()
        return super().cleanup()


class AsyncScheduler(Scheduler):
    """
    Scheduler class to schedule the operations of a thing in an async loop.
    """

    def __init__(self, instance: Thing, rpc_server: RPCServer) -> None:
        super().__init__(instance, rpc_server)
        self._job = None
        self._one_shot = True
        self._operation_execution_ready_event = asyncio.Event()
        self._operation_execution_complete_event = asyncio.Event()

    @property
    def has_job(self) -> bool:
        return self._job is not None

    @property
    def next_job(self) -> Scheduler.JobInvokationType:
        if self._job is None:
            raise RuntimeError("No job to execute")
        return self._job

    def dispatch_job(self, job: Scheduler.JobInvokationType) -> None:
        self._job = job
        eventloop = get_current_async_loop()
        eventloop.call_soon(lambda: asyncio.create_task(self.rpc_server.tunnel_message_to_things(self)))
        eventloop.call_soon(lambda: asyncio.create_task(self.rpc_server.run_thing_instance(self.instance, self)))
        self._job_queued_event.set()


class ThreadedScheduler(Scheduler):
    """
    Scheduler class to schedule the operations of a thing in a threaded loop.
    """

    def __init__(self, instance: Thing, rpc_server: RPCServer) -> None:
        super().__init__(instance, rpc_server)
        self._job = None
        self._execution_thread = None
        self._one_shot = True
        self._operation_execution_ready_event = threading.Event()
        self._operation_execution_complete_event = threading.Event()

    @property
    def has_job(self) -> bool:
        return self._job is not None

    @property
    def next_job(self) -> Scheduler.JobInvokationType:
        if self._job is None:
            raise RuntimeError("No job to execute")
        return self._job

    def dispatch_job(self, job: Scheduler.JobInvokationType) -> None:
        """"""
        self._job = job
        eventloop = get_current_async_loop()
        eventloop.call_soon(lambda: asyncio.create_task(self.rpc_server.tunnel_message_to_things(self)))
        self._execution_thread = threading.Thread(
            target=asyncio.run,
            args=(self.rpc_server.run_thing_instance(self.instance, self),),
        )
        self._execution_thread.start()
        self._job_queued_event.set()


def prepare_rpc_server(
    instance: Thing,
    access_points: str = ZMQ_TRANSPORTS.IPC,
    context: zmq.asyncio.Context | None = None,
    **kwargs,
) -> None:
    # expose_eventloop: bool, False
    #     expose the associated Eventloop which executes the object. This is generally useful for remotely
    #     adding more objects to the same event loop.
    # dont specify http server as a kwarg, as the other method run_with_http_server has to be used
    if context is not None and not isinstance(context, zmq.asyncio.Context):
        raise TypeError("context must be an instance of zmq.asyncio.Context")
    context = context or global_config.zmq_context()

    if access_points == "INPROC" or access_points == ZMQ_TRANSPORTS.INPROC:
        RPCServer(id=instance.id, things=[instance], context=context, logger=instance.logger)
    else:
        from ...server.zmq import ZMQServer

        ZMQServer(
            id=instance.id,
            things=[instance],
            context=context,
            access_points=access_points,
            logger=instance.logger,
        )


__all__ = [RPCServer.__name__]



import unittest
from uuid import UUID, uuid4

from hololinked.core.zmq.message import (
    EXIT,
    OPERATION,
    HANDSHAKE,
    PreserializedData,
    SerializableData,
    RequestHeader,
    EventHeader,
    RequestMessage,
)  # client to server
from hololinked.core.zmq.message import (
    TIMEOUT,
    INVALID_MESSAGE,
    ERROR,
    REPLY,
    ERROR,
    ResponseMessage,
    ResponseHeader,
    EventMessage,
)  # server to client
from hololinked.serializers.serializers import Serializers

try:
    from .utils import TestCase, TestRunner
except ImportError:
    from utils import TestCase, TestRunner


class MessageValidatorMixin(TestCase):
    """A mixin class to validate request and response messages"""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.server_id = f"test-server-{uuid4().hex[:8]}"
        cls.client_id = f"test-client-{uuid4().hex[:8]}"
        cls.thing_id = f"test-thing-{uuid4().hex[:8]}"

    def validate_request_message(self, request_message: RequestMessage) -> None:
        """call this method to validate request message"""

        # req. 1. check message ID is a UUID
        self.assertTrue(isinstance(request_message.id, UUID) or isinstance(UUID(request_message.id, version=4), UUID))
        # req. 2. generated byte array must confine to predefined length (which is readonly & fixed)
        self.assertEqual(len(request_message.byte_array), request_message.length)
        # req. 3. receiver which must be the server ID
        self.assertEqual(request_message.receiver_id, self.server_id)
        # req. 4. sender_id is the client ID
        self.assertEqual(request_message.sender_id, self.client_id)
        # req. 5. all indices of byte array are bytes
        for obj in request_message.byte_array:
            self.assertIsInstance(obj, bytes)
        # req. 6. check that header is correct type (RequestHeader dataclass/struct)
        self.assertIsInstance(request_message.header, RequestHeader)
        # req. 7 check that body is correct type (list of SerializableData and PreserializedData)
        self.assertIsInstance(request_message.body, list)
        self.assertEqual(len(request_message.body), 2)
        self.assertIsInstance(request_message.body[0], SerializableData)
        self.assertIsInstance(request_message.body[1], PreserializedData)

    def validate_response_message(self, response_message: ResponseMessage) -> None:
        """call this method to validate response message"""

        # check message ID is a UUID
        self.assertTrue(isinstance(response_message.id, UUID) or isinstance(UUID(response_message.id, version=4), UUID))
        # check message length
        self.assertEqual(len(response_message.byte_array), response_message.length)
        # check receiver which must be the client
        self.assertEqual(response_message.receiver_id, self.client_id)
        # sender_id is not set before sending message on the socket
        self.assertEqual(response_message.sender_id, self.server_id)
        # check that all indices are bytes
        for obj in response_message.byte_array:
            self.assertIsInstance(obj, bytes)
        # check that header is correct type
        self.assertIsInstance(response_message.header, ResponseHeader)
        # check that body is correct type
        self.assertIsInstance(response_message.body, list)
        self.assertEqual(len(response_message.body), 2)
        self.assertIsInstance(response_message.body[0], SerializableData)
        self.assertIsInstance(response_message.body[1], PreserializedData)

    def validate_event_message(self, event_message: EventMessage) -> None:
        """call this method to validate event message"""

        # check message ID is a UUID
        self.assertTrue(isinstance(event_message.id, UUID) or isinstance(UUID(event_message.id, version=4), UUID))
        # check message length
        self.assertEqual(len(event_message.byte_array), event_message.length)
        # no receiver id for event message, only event id
        self.assertIsInstance(event_message.event_id, str)
        # sender_id is not set before sending message on the socket
        self.assertEqual(event_message.sender_id, self.server_id)
        # check that all indices are bytes
        for obj in event_message.byte_array:
            self.assertIsInstance(obj, bytes)
        # check that header is correct type
        self.assertIsInstance(event_message.header, EventHeader)
        # check that body is correct type
        self.assertIsInstance(event_message.body, list)
        self.assertEqual(len(event_message.body), 2)
        self.assertIsInstance(event_message.body[0], SerializableData)
        self.assertIsInstance(event_message.body[1], PreserializedData)


class TestMessagingContract(MessageValidatorMixin):
    """Tests request and response messages"""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        print(f"test message contract with {cls.__name__}")

    def test_1_request_message(self):
        """test the request message"""

        # request messages types are OPERATION, HANDSHAKE & EXIT
        request_message = RequestMessage.craft_from_arguments(
            receiver_id=self.server_id,
            sender_id=self.client_id,
            thing_id=self.thing_id,
            objekt="some_prop",
            operation="readproperty",
        )
        self.validate_request_message(request_message)
        # check message type for the above craft_from_arguments method
        self.assertEqual(request_message.type, OPERATION)

        request_message = RequestMessage.craft_with_message_type(
            receiver_id=self.server_id, sender_id=self.client_id, message_type=HANDSHAKE
        )
        self.validate_request_message(request_message)
        # check message type for the above craft_with_message_type method
        self.assertEqual(request_message.type, HANDSHAKE)

        request_message = RequestMessage.craft_with_message_type(
            receiver_id=self.server_id, sender_id=self.client_id, message_type=EXIT
        )
        self.validate_request_message(request_message)
        # check message type for the above craft_with_message_type method
        self.assertEqual(request_message.type, EXIT)

    def test_2_response_message(self):
        """test the response message"""

        # response messages types are HANDSHAKE, TIMEOUT, INVALID_MESSAGE, ERROR and REPLY
        response_message = ResponseMessage.craft_from_arguments(
            receiver_id=self.client_id,
            sender_id=self.server_id,
            message_type=HANDSHAKE,
            message_id=uuid4(),
        )
        self.validate_response_message(response_message)
        # check message type for the above craft_with_message_type method
        self.assertEqual(response_message.type, HANDSHAKE)

        response_message = ResponseMessage.craft_from_arguments(
            receiver_id=self.client_id,
            sender_id=self.server_id,
            message_type=TIMEOUT,
            message_id=uuid4(),
        )
        self.validate_response_message(response_message)
        # check message type for the above craft_with_message_type method
        self.assertEqual(response_message.type, TIMEOUT)

        response_message = ResponseMessage.craft_from_arguments(
            receiver_id=self.client_id,
            sender_id=self.server_id,
            message_type=INVALID_MESSAGE,
            message_id=uuid4(),
        )
        self.validate_response_message(response_message)
        # check message type for the above craft_with_message_type method
        self.assertEqual(response_message.type, INVALID_MESSAGE)

        response_message = ResponseMessage.craft_from_arguments(
            receiver_id=self.client_id,
            sender_id=self.server_id,
            message_type=ERROR,
            message_id=uuid4(),
            payload=SerializableData(Exception("test")),
        )
        self.validate_response_message(response_message)
        self.assertEqual(response_message.type, ERROR)
        self.assertIsInstance(Serializers.json.loads(response_message._bytes[2]), dict)

        request_message = RequestMessage.craft_from_arguments(
            sender_id=self.client_id,
            receiver_id=self.server_id,
            thing_id=self.thing_id,
            objekt="some_prop",
            operation="readProperty",
        )
        request_message._sender_id = self.client_id  # will be done by craft_from_self
        response_message = ResponseMessage.craft_reply_from_request(
            request_message=request_message,
        )
        self.validate_response_message(response_message)
        self.assertEqual(response_message.type, REPLY)
        self.assertEqual(Serializers.json.loads(response_message._bytes[3]), None)  # INDEX_BODY = 3
        self.assertEqual(request_message.id, response_message.id)

    def test_3_event_message(self):
        """test the event message"""
        event_message = EventMessage.craft_from_arguments(
            event_id="test-event",
            sender_id=self.server_id,
            payload=SerializableData("test"),
            preserialized_payload=PreserializedData(b"test"),
        )
        self.validate_event_message(event_message)


if __name__ == "__main__":
    unittest.main(testRunner=TestRunner())



import typing
import jsonschema

from ..param.parameterized import Parameterized, ParameterizedMetaclass
from ..constants import JSON
from ..config import global_config


class Event:
    """
    Asynchronously push arbitrary messages to clients. Apart from default events created by the package (like state
    change event, observable properties etc.), events are supposed to be created at class level or at `__init__`
    as a instance attribute, otherwise their publishing socket is unbound and will lead to `AttributeError`.
    """

    # security: Any
    #     security necessary to access this event.

    __slots__ = ["name", "_internal_name", "_publisher", "_observable", "doc", "schema", "security", "label", "owner"]

    def __init__(
        self,
        doc: typing.Optional[str] = None,
        schema: typing.Optional[JSON] = None,
        label: typing.Optional[str] = None,
    ) -> None:
        """
        Parameters
        ----------
        doc: str
            docstring for the event
        schema: JSON
            schema of the event, if the event is JSON complaint. HTTP clients can validate the data with this schema. There
            is no validation on server side.
        label: str
            a descriptive label for the event, to be shown in a GUI for example.
        """

        self.doc = doc
        if global_config.VALIDATE_SCHEMAS and schema:
            jsonschema.Draft7Validator.check_schema(schema)
        self.schema = schema
        # self.security = security
        self.label = label
        self._observable = False

    def __set_name__(self, owner: ParameterizedMetaclass, name: str) -> None:
        self.name = name
        self.owner = owner

    @typing.overload
    def __get__(self, obj, objtype) -> "EventDispatcher": ...

    def __get__(self, obj: Parameterized, objtype: ParameterizedMetaclass = None):
        try:
            if not obj:
                return self
            # uncomment for type hinting
            # from .thing import Thing
            # assert isinstance(obj, Thing)
            return EventDispatcher(
                unique_identifier=f"{obj._qualified_id}/{self.name}",
                publisher=obj.rpc_server.event_publisher if obj.rpc_server else None,
                owner_inst=obj,
                descriptor=self,
            )
        except KeyError:
            raise AttributeError(
                "Event object not yet initialized, please dont access now." + " Access after Thing is running."
            )

    def to_affordance(self, owner_inst=None):
        """
        Generates a `EventAffordance` TD fragment for this Event

        Parameters
        ----------
        owner_inst: Thing, optional
            The instance of the owning `Thing` object. If not supplied, the class is used.

        Returns
        -------
        EventAffordance
            the affordance TD fragment for this event
        """
        from ..td import EventAffordance

        return EventAffordance.generate(self, owner_inst or self.owner)


class EventDispatcher:
    """
    The worker that pushes an event. The separation is necessary between `Event` and
    `EventDispatcher` to allow class level definitions of the `Event`
    """

    __slots__ = ["_unique_identifier", "_publisher", "_owner_inst", "_descriptor"]

    def __init__(
        self, unique_identifier: str, publisher: "EventPublisher", owner_inst: ParameterizedMetaclass, descriptor: Event
    ) -> None:
        self._unique_identifier = unique_identifier
        self._owner_inst = owner_inst
        self._descriptor = descriptor
        self.publisher = publisher

    @property
    def publisher(self) -> "EventPublisher":
        """Event publishing PUB socket owning object"""
        return self._publisher

    @publisher.setter
    def publisher(self, value: "EventPublisher") -> None:
        if not hasattr(self, "_publisher"):
            self._publisher = value
        elif not isinstance(value, EventPublisher):
            raise AttributeError("Publisher must be of type EventPublisher. Given type: " + str(type(value)))
        if self._publisher is not None:
            self._publisher.register(self)

    def push(self, data: typing.Any) -> None:
        """
        publish the event.

        Parameters
        ----------
        data: Any
            payload of the event
        """
        self.publisher.publish(self, data=data)

    def receive_acknowledgement(self, timeout: typing.Union[float, int, None]) -> bool:
        """
        Unimplemented.

        Receive acknowledgement for an event that was just pushed.
        """
        raise NotImplementedError("Event acknowledgement is not implemented yet.")
        return self._synchronize_event.wait(timeout=timeout)

    def _set_acknowledgement(self, *args, **kwargs) -> None:
        """
        Method to be called by RPC server when an acknowledgement is received. Not for user to be set.
        """
        raise NotImplementedError("Event acknowledgement is not implemented yet.")
        self._synchronize_event.set()


from .zmq.brokers import EventPublisher  # noqa: E402

__all__ = [
    Event.__name__,
]



import asyncio
import copy
import threading
import typing
import unittest
import jsonschema
import logging
import random
import time
from types import SimpleNamespace

from hololinked.client.abstractions import SSE
from hololinked.core.actions import BoundAction
from hololinked.core.property import Property
from hololinked.core.thing import Thing
from hololinked.core.zmq.brokers import (
    AsyncEventConsumer,
    AsyncZMQClient,
    EventConsumer,
    EventPublisher,
    SyncZMQClient,
)
from hololinked.core.zmq.message import EXIT, RequestMessage
from hololinked.core.zmq.rpc_server import RPCServer
from hololinked.server.zmq import ZMQServer
from hololinked.td.forms import Form
from hololinked.td.utils import get_zmq_unique_identifier_from_event_affordance
from hololinked.utils import get_all_sub_things_recusively, get_current_async_loop
from hololinked.config import global_config
from hololinked.td import ActionAffordance, PropertyAffordance, EventAffordance
from hololinked.client.zmq.consumed_interactions import ZMQAction, ZMQProperty, ZMQEvent

try:
    from .test_05_brokers import TestBrokerMixin
    from .test_06_actions import replace_methods_with_actions
    from .utils import TestRunner, TestCase
    from .things import (
        run_thing_with_zmq_server_forked,
        test_thing_TD as test_thing_original_TD,
        TestThing,
    )
except ImportError:
    from test_05_brokers import TestBrokerMixin
    from test_06_actions import replace_methods_with_actions
    from utils import TestRunner, TestCase
    from things import (
        run_thing_with_zmq_server_forked,
        test_thing_TD as test_thing_original_TD,
        TestThing,
    )

data_structures = [
    {"key": "value"},
    [1, 2, 3],
    "string",
    42,
    3.14,
    True,
    None,
    {"nested": {"key": "value"}},
    [{"list": "of"}, {"dicts": "here"}],
    {"complex": {"nested": {"list": [1, 2, 3]}, "mixed": [1, "two", 3.0, None]}},
    {"array": [1, 2, 3]},
]  # to use for testing

# global_config.DEBUG = True


class InteractionAffordanceMixin(TestBrokerMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.setUpActions()
        cls.setUpProperties()
        cls.setUpEvents()

    @classmethod
    def setUpActions(cls):
        owner_inst = SimpleNamespace(_noblock_messages={})
        test_thing_TD = copy.deepcopy(test_thing_original_TD)
        test_thing_TD["id"] = cls.thing_id
        cls.action_echo = ZMQAction(
            resource=ActionAffordance.from_TD("action_echo", test_thing_TD),
            sync_client=cls.sync_client,
            async_client=cls.async_client,
            owner_inst=owner_inst,
            logger=cls.logger,
            invokation_timeout=5,
            execution_timeout=5,
        )

        cls.action_get_serialized_data = ZMQAction(
            resource=ActionAffordance.from_TD("get_serialized_data", test_thing_TD),
            sync_client=cls.sync_client,
            async_client=cls.async_client,
            owner_inst=owner_inst,
            logger=cls.logger,
            invokation_timeout=5,
            execution_timeout=5,
        )

        cls.action_sleep = ZMQAction(
            resource=ActionAffordance.from_TD("sleep", test_thing_TD),
            sync_client=cls.sync_client,
            async_client=cls.async_client,
            owner_inst=owner_inst,
            logger=cls.logger,
            invokation_timeout=5,
            execution_timeout=5,
        )

        cls.action_get_mixed_content_data = ZMQAction(
            resource=ActionAffordance.from_TD("get_mixed_content_data", test_thing_TD),
            sync_client=cls.sync_client,
            async_client=cls.async_client,
            owner_inst=owner_inst,
            logger=cls.logger,
            invokation_timeout=5,
            execution_timeout=5,
        )
        cls.action_push_events = ZMQAction(
            resource=ActionAffordance.from_TD("push_events", test_thing_TD),
            sync_client=cls.sync_client,
            async_client=cls.async_client,
            owner_inst=owner_inst,
            logger=cls.logger,
            invokation_timeout=5,
            execution_timeout=5,
        )

    @classmethod
    def setUpProperties(cls):
        owner_inst = SimpleNamespace(_noblock_messages={})
        test_thing_TD = copy.deepcopy(test_thing_original_TD)
        test_thing_TD["id"] = cls.thing_id
        cls.base_property = ZMQProperty(
            resource=PropertyAffordance.from_TD("base_property", test_thing_TD),
            sync_client=cls.sync_client,
            async_client=cls.async_client,
            owner_inst=owner_inst,
            logger=cls.logger,
            invokation_timeout=5,
            execution_timeout=5,
        )
        cls.total_number_of_events = ZMQProperty(
            resource=PropertyAffordance.from_TD("total_number_of_events", test_thing_TD),
            sync_client=cls.sync_client,
            async_client=cls.async_client,
            owner_inst=owner_inst,
            logger=cls.logger,
            invokation_timeout=5,
            execution_timeout=5,
        )

    @classmethod
    def setUpEvents(cls):
        owner_inst = SimpleNamespace(_noblock_messages={})
        test_thing_TD = copy.deepcopy(test_thing_original_TD)
        test_thing_TD["id"] = cls.thing_id
        cls.test_event = ZMQEvent(
            resource=EventAffordance.from_TD("test_event", test_thing_TD),
            owner_inst=owner_inst,
            logger=cls.logger,
        )


class TestRPCServerMixin(InteractionAffordanceMixin):
    @classmethod
    def setUpThing(cls):
        replace_methods_with_actions(TestThing)
        super().setUpThing()

    @classmethod
    def setUpServer(cls):
        cls.server = RPCServer(id=cls.server_id, things=[cls.thing], logger=cls.logger)

    @classmethod
    def setUpClient(cls):
        cls.async_client = AsyncZMQClient(
            id=cls.client_id,
            server_id=cls.server_id,
            access_point="INPROC",
            logger=cls.logger,
            handshake=False,
        )
        cls.sync_client = SyncZMQClient(
            id=cls.client_id + "-sync",
            server_id=cls.server_id,
            access_point="INPROC",
            logger=cls.logger,
            handshake=False,
        )

    @classmethod
    def startServer(cls):
        cls._server_thread = threading.Thread(
            target=cls.server.run,
            daemon=False,  # to test exit daemon must be False
        )
        cls._server_thread.start()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        print(f"test ZMQ RPC Server {cls.__name__}")

    @classmethod
    def tearDownClass(cls):
        cls.server.stop()
        return super().tearDownClass()


class TestInprocRPCServer(TestRPCServerMixin):
    def test_1_creation_defaults(self):
        """test server configuration defaults"""
        self.assertTrue(self.server.req_rep_server.socket_address.startswith("inproc://"))
        self.assertTrue(self.server.event_publisher.socket_address.startswith("inproc://"))

        self.assertTrue(self.thing.rpc_server, self.server)
        self.assertTrue(self.thing.event_publisher, self.server.event_publisher)

    def test_2_handshake(self):
        """test handshake mechanisms"""
        self.sync_client.handshake()

        async def async_handshake():
            self.async_client.handshake()
            await self.async_client.handshake_complete()

        get_current_async_loop().run_until_complete(async_handshake())

    def test_3_action_abstractions(self):
        """ "test if action can be invoked by a client"""

        async def test_basic_operations():
            """Test if action can be invoked by a client in basic request/response way, oneway and no block"""
            nonlocal self
            await self.action_echo.async_call("value")
            self.action_echo.oneway(5)
            noblock_msg_id = self.action_echo.noblock(10)
            self.assertEqual(self.action_echo.last_return_value, "value")
            # test the responses for no block call, so read the socket - but, this is usually abstracte in a higher level API
            response = self.action_echo._sync_zmq_client.recv_response(noblock_msg_id)
            self.action_echo._last_zmq_response = response
            self.assertEqual(self.action_echo.last_return_value, 10)
            self.assertEqual(self.action_echo(2), 2)

        get_current_async_loop().run_until_complete(test_basic_operations())
        self.sync_client.handshake()

        async def test_operations_thorough():
            # Generate 20 random JSON serializable data structures
            nonlocal self
            global data_structures

            msg_ids = [None for i in range(len(data_structures))]
            last_call_type = None
            # Randomize calls to self.action_echo
            for index, data in enumerate(data_structures):
                call_type = random.choice(["async_call", "plain_call", "oneway", "noblock"])
                if call_type == "async_call":
                    result = await self.action_echo.async_call(data)
                    self.assertEqual(result, data)
                elif call_type == "plain_call":
                    result = self.action_echo(data)
                    self.assertEqual(result, data)
                elif call_type == "oneway":
                    self.action_echo.oneway(data)
                    self.assertNotEqual(data, self.action_echo.last_return_value)
                elif call_type == "noblock":
                    msg_ids[index] = self.action_echo.noblock(data)
                    self.assertNotEqual(data, self.action_echo.last_return_value)

                # print("last_call_type", last_call_type, "call_type", call_type, "data", data)
                if last_call_type == "noblock":
                    response = self.action_echo._sync_zmq_client.recv_response(msg_ids[index - 1])
                    self.action_echo._last_zmq_response = response
                    self.assertEqual(self.action_echo.last_return_value, data_structures[index - 1])

                last_call_type = call_type

        get_current_async_loop().run_until_complete(test_operations_thorough())
        self.sync_client.handshake()

    def test_4_property_abstractions(self):
        """Test if property can be invoked by a client"""

        def test_basic_operations():
            nonlocal self
            self.base_property.set(100)
            self.assertEqual(self.base_property.get(), 100)
            self.base_property.oneway_set(200)
            self.assertEqual(self.base_property.get(), 200)

            async def test_async_property_abstractions():
                nonlocal self
                await self.base_property.async_set(300)
                self.assertEqual(self.base_property.get(), 300)
                await self.base_property.async_set(0)
                self.assertEqual(await self.base_property.async_get(), 0)

            get_current_async_loop().run_until_complete(test_async_property_abstractions())

        test_basic_operations()
        self.sync_client.handshake()

        async def test_operations_thorough():
            # Generate 20 random JSON serializable data structures
            nonlocal self
            global data_structures

            msg_ids = [None for i in range(len(data_structures))]
            last_call_type = None
            # Randomize calls to self.action_echo
            for index, data in enumerate(data_structures):
                call_type = random.choice(["async_set", "set", "oneway_set", "noblock_get"])
                if call_type == "async_set":
                    self.assertIsNone(await self.base_property.async_set(data))
                    self.assertEqual(await self.base_property.async_get(), data)
                elif call_type == "set":
                    self.assertIsNone(self.base_property.set(data))
                    self.assertEqual(self.base_property.get(), data)
                elif call_type == "oneway_set":
                    self.assertIsNone(self.base_property.oneway_set(data))
                    self.assertNotEqual(data, self.base_property.last_read_value)
                    self.assertEqual(data, self.base_property.get())
                    # for one way calls as well, get() will return the latest value
                elif call_type == "noblock_get":
                    msg_ids[index] = self.base_property.noblock_get()
                    self.assertNotEqual(data, self.base_property.last_read_value)

                #  print("last_call_type", last_call_type, "call_type", call_type, "data", data)
                if last_call_type == "noblock":
                    response = self.base_property._sync_zmq_client.recv_response(msg_ids[index - 1])
                    self.base_property._last_zmq_response = response
                    self.assertEqual(self.base_property.last_read_value, data_structures[index - 1])

                last_call_type = call_type

        get_current_async_loop().run_until_complete(test_operations_thorough())
        self.sync_client.handshake()

    def test_5_thing_execution_context(self):
        """test if thing execution context is used correctly"""
        old_thing_execution_context = self.action_echo._thing_execution_context
        # Only fetch_execution_logs currently supported
        self.action_echo._thing_execution_context = dict(fetch_execution_logs=True)
        get_current_async_loop().run_until_complete(self.action_echo.async_call("value"))
        self.assertIsInstance(self.action_echo.last_return_value, dict)
        self.assertTrue("execution_logs" in self.action_echo.last_return_value.keys())
        self.assertTrue("return_value" in self.action_echo.last_return_value.keys())
        self.assertTrue(len(self.action_echo.last_return_value) == 2)
        self.assertFalse(self.action_echo.last_return_value == "value")  # because its a dict now
        self.assertIsInstance(self.action_echo.last_return_value["execution_logs"], list)
        self.assertTrue(self.action_echo.last_return_value["return_value"] == "value")
        self.action_echo._thing_execution_context = old_thing_execution_context

    def test_6_server_execution_context(self):
        """test if server execution context is used correctly"""

        async def test_execution_timeout():
            try:
                await self.action_sleep.async_call()
            except Exception as ex:
                self.assertIsInstance(ex, TimeoutError)
                self.assertIn("Execution timeout occured", str(ex))
            else:
                self.assertTrue(False)  # fail the test if reached here

        get_current_async_loop().run_until_complete(test_execution_timeout())

        async def test_invokation_timeout():
            try:
                old_timeout = self.action_sleep._invokation_timeout
                self.action_sleep._invokation_timeout = 0.1  # reduce the value to test timeout
                await self.action_sleep.async_call()
            except Exception as ex:
                self.assertIsInstance(ex, TimeoutError)
                self.assertIn("Invokation timeout occured", str(ex))
            else:
                self.assertTrue(False)  # fail the test if reached here
            finally:
                self.action_sleep._invokation_timeout = old_timeout

        get_current_async_loop().run_until_complete(test_invokation_timeout())

    def test_7_binary_payloads(self):
        """test if binary payloads are handled correctly"""
        self.assertEqual(self.action_get_mixed_content_data(), ("foobar", b"foobar"))
        self.assertEqual(self.action_get_serialized_data(), b"foobar")

        async def async_call():
            await self.action_get_mixed_content_data.async_call()
            return self.action_get_mixed_content_data.last_return_value

        result = get_current_async_loop().run_until_complete(async_call())
        self.assertEqual(result, ("foobar", b"foobar"))

        async def async_call():
            await self.action_get_serialized_data.async_call()
            return self.action_get_serialized_data.last_return_value

        result = get_current_async_loop().run_until_complete(async_call())
        self.assertEqual(result, b"foobar")

    def test_8_stop(self):
        """test if server can be stopped"""
        self.server.stop()


class TestRPCServer(TestInprocRPCServer):
    @classmethod
    def setUpServer(cls):
        cls.server = ZMQServer(
            id=cls.server_id,
            things=[cls.thing],
            access_points=["INPROC", "IPC", "tcp://*:59000"],
            logger=cls.logger,
        )

    @classmethod
    def setUpClient(cls):
        super().setUpClient()
        cls.sync_ipc_client = SyncZMQClient(
            id=cls.client_id + "-sync",
            server_id=cls.server_id,
            logger=cls.logger,
            handshake=False,
            access_point="IPC",
        )
        cls.sync_tcp_client = SyncZMQClient(
            id=cls.client_id + "-sync",
            server_id=cls.server_id,
            logger=cls.logger,
            handshake=False,
            access_point="tcp://localhost:59000",
        )
        cls.async_ipc_client = AsyncZMQClient(
            id=cls.client_id + "-async",
            server_id=cls.server_id,
            logger=cls.logger,
            handshake=False,
            access_point="IPC",
        )
        cls.async_tcp_client = AsyncZMQClient(
            id=cls.client_id + "-async",
            server_id=cls.server_id,
            logger=cls.logger,
            handshake=False,
            access_point="tcp://localhost:59000",
        )

    def test_1_creation_defaults(self):
        super().test_1_creation_defaults()
        # check socket creation defaults
        self.assertTrue(self.server.ipc_server.socket_address.startswith("ipc://"))
        self.assertTrue(self.server.tcp_server.socket_address.startswith("tcp://"))
        self.assertTrue(self.server.tcp_server.socket_address.endswith(":59000"))

    def test_2_handshake(self):
        super().test_2_handshake()
        self.sync_ipc_client.handshake()
        self.sync_tcp_client.handshake()

        async def async_handshake():
            self.async_ipc_client.handshake()
            await self.async_ipc_client.handshake_complete()
            self.async_tcp_client.handshake()
            await self.async_tcp_client.handshake_complete()

        get_current_async_loop().run_until_complete(async_handshake())

    def test_3_action_abstractions(self):
        old_sync_client = self.action_echo._sync_zmq_client
        old_async_client = self.action_echo._async_zmq_client
        for clients in [
            (self.sync_tcp_client, self.async_tcp_client),
            (self.sync_ipc_client, self.async_ipc_client),
        ]:
            self.action_echo._sync_zmq_client, self.action_echo._async_zmq_client = clients
            super().test_3_action_abstractions()
        self.action_echo._sync_zmq_client = old_sync_client
        self.action_echo._async_zmq_client = old_async_client

    def test_4_property_abstractions(self):
        old_sync_client = self.base_property._sync_zmq_client
        old_async_client = self.base_property._async_zmq_client
        for clients in [
            (self.sync_tcp_client, self.async_tcp_client),
            (self.sync_ipc_client, self.async_ipc_client),
        ]:
            (
                self.base_property._sync_zmq_client,
                self.base_property._async_zmq_client,
            ) = clients
            super().test_4_property_abstractions()
        self.base_property._sync_zmq_client = old_sync_client
        self.base_property._async_zmq_client = old_async_client

    def test_5_thing_execution_context(self):
        old_sync_client = self.action_echo._sync_zmq_client
        old_async_client = self.action_echo._async_zmq_client
        for clients in [
            (self.sync_tcp_client, self.async_tcp_client),
            (self.sync_ipc_client, self.async_ipc_client),
        ]:
            self.action_echo._sync_zmq_client, self.action_echo._async_zmq_client = clients
            super().test_5_thing_execution_context()
        self.action_echo._sync_zmq_client = old_sync_client
        self.action_echo._async_zmq_client = old_async_client

    def test_6_server_execution_context(self):
        old_sync_client = self.action_sleep._sync_zmq_client
        old_async_client = self.action_sleep._async_zmq_client
        for clients in [
            (self.sync_tcp_client, self.async_tcp_client),
            (self.sync_ipc_client, self.async_ipc_client),
        ]:
            self.action_sleep._sync_zmq_client, self.action_sleep._async_zmq_client = clients
            super().test_6_server_execution_context()
        self.action_sleep._sync_zmq_client = old_sync_client
        self.action_sleep._async_zmq_client = old_async_client

    def test_7_binary_payloads(self):
        for clients in [
            (self.sync_tcp_client, self.async_tcp_client),
            (self.sync_ipc_client, self.async_ipc_client),
        ]:
            for action in [
                self.action_get_serialized_data,
                self.action_get_mixed_content_data,
            ]:
                action._sync_zmq_client, action._async_zmq_client = clients
            super().test_7_binary_payloads()


class TestExposedActions(InteractionAffordanceMixin):
    @classmethod
    def setUpServer(cls):
        pass

    @classmethod
    def setUpThing(cls):
        pass

    @classmethod
    def startServer(cls):
        run_thing_with_zmq_server_forked(
            thing_cls=TestThing,
            id=cls.server_id,
            log_level=logging.ERROR + 10,
            done_queue=cls.done_queue,
            prerun_callback=replace_methods_with_actions,
            as_process=False,
        )

    @classmethod
    def setUpClient(cls):
        super().setUpClient()
        cls.sync_client = SyncZMQClient(
            id=cls.client_id,
            server_id=cls.server_id,
            logger=cls.logger,
            handshake=False,
        )
        cls.client = cls.sync_client

    def test_1_exposed_actions(self):
        """
        Now that actions can be invoked by a client, test different types of actions
        and their behaviors
        """
        replace_methods_with_actions(TestThing)
        thing = TestThing(id=self.server_id, log_level=logging.ERROR)
        # has to match server only because run_thing_with_zmq_server_forked equates server_id and thing_id
        self.sync_client.handshake()

        # thing_client = ObjectProxy('test-action', log_level=logging.ERROR) # type: TestThing
        assert isinstance(thing.action_echo, BoundAction)  # type definition
        action_echo = ZMQAction(
            resource=thing.action_echo.to_affordance(),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        self.assertEqual(action_echo(1), 1)

        assert isinstance(thing.action_echo_with_classmethod, BoundAction)  # type definition
        action_echo_with_classmethod = ZMQAction(
            resource=thing.action_echo_with_classmethod.to_affordance(),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        self.assertEqual(action_echo_with_classmethod(2), 2)

        assert isinstance(thing.action_echo_async, BoundAction)  # type definition
        action_echo_async = ZMQAction(
            resource=thing.action_echo_async.to_affordance(),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        self.assertEqual(action_echo_async("string"), "string")

        assert isinstance(thing.action_echo_async_with_classmethod, BoundAction)  # type definition
        action_echo_async_with_classmethod = ZMQAction(
            resource=thing.action_echo_async_with_classmethod.to_affordance(),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        self.assertEqual(action_echo_async_with_classmethod([1, 2]), [1, 2])

        assert isinstance(thing.parameterized_action, BoundAction)  # type definition
        parameterized_action = ZMQAction(
            resource=thing.parameterized_action.to_affordance(),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        self.assertEqual(
            parameterized_action(arg1=1, arg2="hello", arg3=5),
            [self.server_id, 1, "hello", 5],
        )

        assert isinstance(thing.parameterized_action_async, BoundAction)  # type definition
        parameterized_action_async = ZMQAction(
            resource=thing.parameterized_action_async.to_affordance(),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        self.assertEqual(
            parameterized_action_async(arg1=2.5, arg2="hello", arg3="foo"),
            [self.server_id, 2.5, "hello", "foo"],
        )

        assert isinstance(thing.parameterized_action_without_call, BoundAction)  # type definition
        parameterized_action_without_call = ZMQAction(
            resource=thing.parameterized_action_without_call.to_affordance(),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        with self.assertRaises(NotImplementedError) as ex:
            parameterized_action_without_call(arg1=2, arg2="hello", arg3=5)
        self.assertTrue(str(ex.exception).startswith("Subclasses must implement __call__"))

    def test_2_schema_validation(self):
        """Test if schema validation is working correctly"""
        self._test_2_json_schema_validation()
        self._test_2_pydantic_validation()

    def _test_2_json_schema_validation(self):
        thing = TestThing(id=self.server_id, log_level=logging.ERROR)
        self.sync_client.handshake()

        # JSON schema validation
        assert isinstance(thing.json_schema_validated_action, BoundAction)  # type definition
        action_affordance = thing.json_schema_validated_action.to_affordance()
        json_schema_validated_action = ZMQAction(
            resource=action_affordance, sync_client=self.client, async_client=None, logger=self.logger, owner_inst=None
        )
        # data with invalid schema
        with self.assertRaises(Exception) as ex1:
            json_schema_validated_action(val1="1", val2="hello", val3={"field": "value"}, val4=[])
        self.assertTrue(str(ex1.exception).startswith("'1' is not of type 'integer'"))
        with self.assertRaises(Exception) as ex2:
            json_schema_validated_action("1", val2="hello", val3={"field": "value"}, val4=[])
        self.assertTrue(str(ex2.exception).startswith("'1' is not of type 'integer'"))
        with self.assertRaises(Exception) as ex3:
            json_schema_validated_action(1, 2, val3={"field": "value"}, val4=[])
        self.assertTrue(str(ex3.exception).startswith("2 is not of type 'string'"))
        with self.assertRaises(Exception) as ex4:
            json_schema_validated_action(1, "hello", val3="field", val4=[])
        self.assertTrue(str(ex4.exception).startswith("'field' is not of type 'object'"))
        with self.assertRaises(Exception) as ex5:
            json_schema_validated_action(1, "hello", val3={"field": "value"}, val4="[]")
        self.assertTrue(str(ex5.exception).startswith("'[]' is not of type 'array'"))
        # data with valid schema
        return_value = json_schema_validated_action(val1=1, val2="hello", val3={"field": "value"}, val4=[])
        self.assertEqual(return_value, {"val1": 1, "val3": {"field": "value"}})
        jsonschema.Draft7Validator(action_affordance.output).validate(return_value)

    def _test_2_pydantic_validation(self):
        thing = TestThing(id=self.server_id, log_level=logging.ERROR)
        self.sync_client.handshake()

        # Pydantic schema validation
        assert isinstance(thing.pydantic_validated_action, BoundAction)  # type definition
        action_affordance = thing.pydantic_validated_action.to_affordance()
        pydantic_validated_action = ZMQAction(
            resource=action_affordance, sync_client=self.client, async_client=None, logger=self.logger, owner_inst=None
        )
        # data with invalid schema
        with self.assertRaises(Exception) as ex1:
            pydantic_validated_action(val1="1", val2="hello", val3={"field": "value"}, val4=[])
        self.assertTrue(
            "validation error for pydantic_validated_action_input" in str(ex1.exception)
            and "val1" in str(ex1.exception)
            and "val2" not in str(ex1.exception)
            and "val3" not in str(ex1.exception)
            and "val4" not in str(ex1.exception)
        )  # {obj.name}_input is the pydantic model name
        with self.assertRaises(Exception) as ex2:
            pydantic_validated_action("1", val2="hello", val3={"field": "value"}, val4=[])
        self.assertTrue(
            "validation error for pydantic_validated_action_input" in str(ex2.exception)
            and "val1" in str(ex2.exception)
            and "val2" not in str(ex2.exception)
            and "val3" not in str(ex2.exception)
            and "val4" not in str(ex2.exception)
        )
        with self.assertRaises(Exception) as ex3:
            pydantic_validated_action(1, 2, val3={"field": "value"}, val4=[])
        self.assertTrue(
            "validation error for pydantic_validated_action_input" in str(ex3.exception)
            and "val1" not in str(ex3.exception)
            and "val2" in str(ex3.exception)
            and "val3" not in str(ex3.exception)
            and "val4" not in str(ex3.exception)
        )
        with self.assertRaises(Exception) as ex4:
            pydantic_validated_action(1, "hello", val3="field", val4=[])
        self.assertTrue(
            "validation error for pydantic_validated_action_input" in str(ex4.exception)
            and "val1" not in str(ex4.exception)
            and "val2" not in str(ex4.exception)
            and "val3" in str(ex4.exception)
            and "val4" not in str(ex4.exception)
        )
        with self.assertRaises(Exception) as ex5:
            pydantic_validated_action(1, "hello", val3={"field": "value"}, val4="[]")
        self.assertTrue(
            "validation error for pydantic_validated_action_input" in str(ex5.exception)
            and "val1" not in str(ex5.exception)
            and "val2" not in str(ex5.exception)
            and "val3" not in str(ex5.exception)
            and "val4" in str(ex5.exception)
        )
        # data with valid schema
        return_value = pydantic_validated_action(val1=1, val2="hello", val3={"field": "value"}, val4=[])
        self.assertEqual(return_value, {"val2": "hello", "val4": []})

    def test_3_exit(self):
        """Exit the server"""
        exit_message = RequestMessage.craft_with_message_type(
            sender_id="test-action-client",
            receiver_id=self.server_id,
            message_type=EXIT,
        )
        self.sync_client.socket.send_multipart(exit_message.byte_array)
        self.assertEqual(self.done_queue.get(), self.server_id)


class TestExposedProperties(InteractionAffordanceMixin):
    @classmethod
    def setUpThing(cls):
        pass

    @classmethod
    def setUpServer(cls):
        pass

    @classmethod
    def startServer(cls):
        run_thing_with_zmq_server_forked(
            thing_cls=TestThing,
            id=cls.server_id,
            log_level=logging.ERROR + 10,
            done_queue=cls.done_queue,
            as_process=False,
        )

    @classmethod
    def setUpClient(cls):
        super().setUpClient()
        cls.sync_client = SyncZMQClient(
            id=cls.client_id,
            server_id=cls.server_id,
            logger=cls.logger,
            handshake=False,
        )
        cls.client = cls.sync_client

    def test_01_property_abstractions(self):
        thing = TestThing(id=self.server_id, log_level=logging.ERROR)
        self.sync_client.handshake()

        descriptor = thing.properties["number_prop"]
        assert isinstance(descriptor, Property)  # type definition
        number_prop = ZMQProperty(
            resource=descriptor.to_affordance(thing),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        self.assertEqual(number_prop.get(), descriptor.default)
        number_prop.set(100)
        self.assertEqual(number_prop.get(), 100)
        number_prop.oneway_set(200)
        self.assertEqual(number_prop.get(), 200)

        async def test_6_async_property_abstractions(self: "TestThing"):
            nonlocal number_prop
            async_client = AsyncZMQClient(
                id="test-property-async-client",
                server_id=self.server_id,
                log_level=logging.ERROR,
                handshake=False,
            )
            number_prop._async_zmq_client = async_client
            async_client.handshake()
            await async_client.handshake_complete()
            await number_prop.async_set(300)
            self.assertEqual(number_prop.get(), 300)
            await number_prop.async_set(0)
            self.assertEqual(await number_prop.async_get(), 0)

        get_current_async_loop().run_until_complete(test_6_async_property_abstractions(self))

    def test_02_json_schema_property(self):
        """Test json schema based property"""
        test_thing_TD = copy.deepcopy(test_thing_original_TD)
        test_thing_TD["id"] = self.server_id  # thing id should be server id. TODO refactor this
        json_schema_prop = ZMQProperty(
            resource=PropertyAffordance.from_TD("json_schema_prop", test_thing_TD),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        json_schema_prop.resource._thing_id = self.server_id
        json_schema_prop.set("hello")
        self.assertEqual(json_schema_prop.get(), "hello")
        json_schema_prop.set("world")
        self.assertEqual(json_schema_prop.get(), "world")
        with self.assertRaises(Exception) as ex:
            json_schema_prop.set("world1")
        self.assertTrue("Failed validating 'pattern' in schema:" in str(ex.exception))

    def test_03_pydantic_model_property(self):
        """Test pydantic model based property"""
        test_thing_TD = copy.deepcopy(test_thing_original_TD)
        test_thing_TD["id"] = self.server_id  # thing id should be server id. TODO refactor this
        pydantic_prop = ZMQProperty(
            resource=PropertyAffordance.from_TD("pydantic_prop", test_thing_TD),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        pydantic_prop.resource._thing_id = self.server_id

        valid_value = {"foo": "foo", "bar": 1, "foo_bar": 1.0}
        pydantic_prop.set(valid_value)
        self.assertEqual(pydantic_prop.get(), valid_value)

        invalid_value = {"foo": 1, "bar": "1", "foo_bar": 1.0}
        with self.assertRaises(Exception) as ex:
            pydantic_prop.set(invalid_value)
        self.assertTrue("validation error for PydanticProp" in str(ex.exception))

        pydantic_simple_prop = ZMQProperty(
            resource=PropertyAffordance.from_TD("pydantic_simple_prop", test_thing_TD),
            sync_client=self.client,
            async_client=None,
            logger=self.logger,
            owner_inst=None,
        )
        pydantic_simple_prop.resource._thing_id = self.server_id
        pydantic_simple_prop.set(5)
        self.assertEqual(pydantic_simple_prop.get(), 5)
        with self.assertRaises(Exception) as ex:
            pydantic_simple_prop.set("5str")
        self.assertTrue("validation error for 'int'" in str(ex.exception))

    def test_04_exit(self):
        exit_message = RequestMessage.craft_with_message_type(
            sender_id="test-property-client",
            receiver_id=self.server_id,
            message_type=EXIT,
        )
        self.sync_client.socket.send_multipart(exit_message.byte_array)
        self.assertEqual(self.done_queue.get(), self.server_id)


class TestExposedEvents(TestRPCServerMixin):
    @classmethod
    def setUpServer(cls):
        cls.server = ZMQServer(
            id=cls.server_id,
            things=[cls.thing],
            logger=cls.logger,
            access_points=["INPROC", "IPC", "tcp://*:59005"],
        )

    @classmethod
    def setUpEvents(cls):
        test_thing_TD = copy.deepcopy(test_thing_original_TD)
        test_thing_TD["id"] = cls.thing_id
        cls.event_names = [
            "test_event",
            "test_binary_payload_event",
            "test_event_with_json_schema",
        ]
        for event_name in cls.event_names:
            event_affordance = EventAffordance.from_TD(event_name, test_thing_TD)
            form = Form()
            form.href = cls.server.event_publisher.socket_address
            form.contentType = "application/json"
            form.op = "subscribeevent"
            form.subprotocol = "sse"
            event_affordance.forms = [form]
            event = ZMQEvent(resource=event_affordance, logger=cls.logger, owner_inst=None)
            setattr(cls, event_name, event)

    def test_1_creation_defaults(self):
        """test server configuration defaults"""
        all_things = get_all_sub_things_recusively(self.thing)
        self.assertTrue(len(all_things) > 1)  # run the test only if there are sub things
        for thing in all_things:
            assert isinstance(thing, Thing)
            for name, event in thing.events.values.items():
                self.assertTrue(event.publisher, self.server.event_publisher)
                self.assertIsInstance(event._unique_identifier, str)
                self.assertEqual(event._owner_inst, thing)

    def test_2_sync_client_event_stream(self):
        """test if event can be streamed by a synchronous threaded client"""

        def test_events(event_name: str, expected_data: typing.Any) -> None:
            event_client = getattr(self, event_name)  # type: ZMQEvent

            self.assertEqual(
                get_zmq_unique_identifier_from_event_affordance(event_client.resource),
                getattr(self.thing, event_client.resource.name)._unique_identifier,  # type: EventDispatcher
            )
            attempts = 100
            results = []

            def cb(value: SSE):
                nonlocal results
                results.append(value)

            event_client.subscribe(cb)
            time.sleep(5)  # calm down for event publisher to connect fully as there is no handshake for events
            self.action_push_events(event_name=event_name, total_number_of_events=attempts)

            for i in range(attempts):
                if len(results) == attempts:
                    break
                time.sleep(0.1)
            self.assertAlmostEqual(len(results), attempts, delta=3)
            self.assertEqual([res.data for res in results], [expected_data] * len(results))
            event_client.unsubscribe()

        for name, data in zip(
            self.event_names,
            [
                "test data",
                b"test data",
                {
                    "val1": 1,
                    "val2": "test",
                    "val3": {"key": "value"},
                    "val4": [1, 2, 3],
                },
            ],
        ):
            test_events(name, data)

    def test_3_async_client_event_stream(self):
        """test if event can be streamed by an asynchronous client in an async loop"""

        async def test_events(event_name: str, expected_data: typing.Any) -> None:
            event_client = getattr(self, event_name)  # type: ZMQEvent
            self.assertEqual(
                get_zmq_unique_identifier_from_event_affordance(event_client.resource),
                getattr(self.thing, event_client.resource.name)._unique_identifier,  # type: EventDispatcher
            )
            attempts = 100
            results = []

            def cb(value: SSE):
                nonlocal results
                # print("event callback", value)
                results.append(value)

            event_client.subscribe(cb, asynch=True)
            time.sleep(5)  # calm down for event publisher to connect fully as there is no handshake for events
            self.action_push_events(event_name=event_name, total_number_of_events=attempts)

            for i in range(attempts):
                if len(results) == attempts:
                    break
                await asyncio.sleep(0.1)
            self.assertAlmostEqual(len(results), attempts, delta=3)
            # since we are pushing events in multiple protocols, sometimes the event from the previous test is
            # still lingering on the socket. So the captured event must be at least the number of attempts.
            self.assertEqual([res.data for res in results], [expected_data] * len(results))
            event_client.unsubscribe()

        for name, data in zip(
            self.event_names,
            [
                "test data",
                b"test data",
                {
                    "val1": 1,
                    "val2": "test",
                    "val3": {"key": "value"},
                    "val4": [1, 2, 3],
                },
            ],
        ):
            get_current_async_loop().run_until_complete(test_events(name, data))

    def test_4_other_transports(self):
        test_thing_TD = copy.deepcopy(test_thing_original_TD)
        test_thing_TD["id"] = self.thing_id
        for publisher in [
            self.server.ipc_event_publisher,
            self.server.tcp_event_publisher,
        ]:
            self.assertIsInstance(publisher, EventPublisher)
            self.assertTrue(
                publisher.socket_address.startswith("tcp://") or publisher.socket_address.startswith("ipc://")
            )
            for event_name in self.event_names:
                event_affordance = EventAffordance.from_TD(event_name, test_thing_TD)
                event = getattr(self, event_name)  # type: ZMQEvent
                form = Form()
                form.href = publisher.socket_address.replace("*", "localhost")
                form.contentType = "application/json"
                form.op = "subscribeevent"
                form.subprotocol = "sse"
                event_affordance.forms = [form]
                event.resource.forms = event_affordance.forms
            self.test_2_sync_client_event_stream()
            self.test_3_async_client_event_stream()

    def test_5_exit(self):
        self.server.stop()


class TestThingRunRPCServer(TestBrokerMixin):
    """Finally check if the thing can be run with a ZMQ server directly"""

    @classmethod
    def setUpThing(self):
        self.thing = TestThing(id=self.thing_id, logger=self.logger, remote_accessible_logger=True)

    @classmethod
    def startServer(self):
        self.thing.run_with_zmq_server(forked=True)
        self.server = self.thing.rpc_server
        self.sync_client = SyncZMQClient(
            id=self.client_id,
            server_id=self.thing_id,
            logger=self.logger,
            handshake=False,
            access_point="INPROC",
        )
        self.async_client = AsyncZMQClient(
            id=self.client_id + "async",
            server_id=self.thing_id,
            logger=self.logger,
            handshake=False,
            access_point="INPROC",
        )
        time.sleep(2)

    def test_1_setup_zmq_server(self):
        self.assertIsInstance(self.thing.rpc_server, ZMQServer)
        self.assertIsInstance(self.thing.event_publisher, EventPublisher)

    def test_2_handshake(self):
        self.sync_client.handshake()
        self.async_client.handshake()
        get_current_async_loop().run_until_complete(self.async_client.handshake_complete())

    def test_3_stop(self):
        self.thing.rpc_server.stop()


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestInprocRPCServer))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestRPCServer))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestExposedActions))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestExposedProperties))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestExposedEvents))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestThingRunRPCServer))
    return suite


if __name__ == "__main__":
    runner = TestRunner()
    runner.run(load_tests(unittest.TestLoader(), None, None))



import typing
from .base import Schema


class SecurityScheme(Schema):
    """
    create security scheme.
    schema - https://www.w3.org/TR/wot-thing-description11/#sec-security-vocabulary-definition
    """

    scheme: str = None
    description: str = None
    descriptions: typing.Optional[typing.Dict[str, str]] = None
    proxy: typing.Optional[str] = None

    def __init__(self):
        super().__init__()

    def build(self):
        self.scheme = "nosec"
        self.description = (
            "currently no security scheme supported - use cookie auth directly on hololinked.server.HTTPServer object"
        )



import logging
import unittest
import tempfile
import os
import json
import copy
import pydantic

from hololinked.core.properties import Number
from hololinked.storage.database import BaseDB, ThingDB
from hololinked.serializers import PythonBuiltinJSONSerializer

try:
    from .utils import TestCase, TestRunner
    from .things import TestThing
except ImportError:
    from utils import TestCase, TestRunner
    from things import TestThing


class TestProperty(TestCase):
    def test_01_simple_class_property(self):
        """Test basic class property functionality"""
        # Test class-level access
        self.assertEqual(TestThing.simple_class_prop, 42)
        TestThing.simple_class_prop = 100
        self.assertEqual(TestThing.simple_class_prop, 100)

        # Test that instance-level access reflects class value
        instance1 = TestThing(id="test1", log_level=logging.WARN)
        instance2 = TestThing(id="test2", log_level=logging.WARN)
        self.assertEqual(instance1.simple_class_prop, 100)
        self.assertEqual(instance2.simple_class_prop, 100)

        # Test that instance-level changes affect class value
        instance1.simple_class_prop = 200
        self.assertEqual(TestThing.simple_class_prop, 200)
        self.assertEqual(instance2.simple_class_prop, 200)

    def test_02_managed_class_property(self):
        """Test class property with custom getter/setter"""
        # Test initial value
        self.assertEqual(TestThing.managed_class_prop, 0)
        # Test valid value assignment
        TestThing.managed_class_prop = 50
        self.assertEqual(TestThing.managed_class_prop, 50)
        # Test validation in setter
        with self.assertRaises(ValueError):
            TestThing.managed_class_prop = -10
        # Verify value wasn't changed after failed assignment
        self.assertEqual(TestThing.managed_class_prop, 50)
        # Test instance-level validation
        instance = TestThing(id="test3", log_level=logging.WARN)
        with self.assertRaises(ValueError):
            instance.managed_class_prop = -20
        # Test that instance-level access reflects class value
        self.assertEqual(instance.managed_class_prop, 50)
        # Test that instance-level changes affects class value
        instance.managed_class_prop = 100
        self.assertEqual(TestThing.managed_class_prop, 100)
        self.assertEqual(instance.managed_class_prop, 100)

    def test_03_readonly_class_property(self):
        """Test read-only class property behavior"""
        # Test reading the value
        self.assertEqual(TestThing.readonly_class_prop, "read-only-value")

        # Test that setting raises an error at class level
        with self.assertRaises(ValueError):
            TestThing.readonly_class_prop = "new-value"

        # Test that setting raises an error at instance level
        instance = TestThing(id="test4", log_level=logging.WARN)
        with self.assertRaises(ValueError):
            instance.readonly_class_prop = "new-value"

        # Verify value remains unchanged
        self.assertEqual(TestThing.readonly_class_prop, "read-only-value")
        self.assertEqual(instance.readonly_class_prop, "read-only-value")

    def test_04_deletable_class_property(self):
        """Test class property deletion"""
        # Test initial value
        self.assertEqual(TestThing.deletable_class_prop, 100)

        # Test setting new value
        TestThing.deletable_class_prop = 150
        self.assertEqual(TestThing.deletable_class_prop, 150)

        # Test deletion
        instance = TestThing(id="test5", log_level=logging.WARN)
        del TestThing.deletable_class_prop
        self.assertEqual(TestThing.deletable_class_prop, 100)  # Should return to default
        self.assertEqual(instance.deletable_class_prop, 100)

        # Test instance-level deletion
        instance.deletable_class_prop = 200
        self.assertEqual(TestThing.deletable_class_prop, 200)
        del instance.deletable_class_prop
        self.assertEqual(TestThing.deletable_class_prop, 100)  # Should return to default

    def test_05_descriptor_access(self):
        """Test descriptor access for class properties"""
        # Test direct access through descriptor
        instance = TestThing(id="test6", log_level=logging.WARN)
        self.assertIsInstance(TestThing.not_a_class_prop, Number)
        self.assertEqual(instance.not_a_class_prop, 43)
        instance.not_a_class_prop = 50
        self.assertEqual(instance.not_a_class_prop, 50)

        del instance.not_a_class_prop
        # deleter deletes only an internal instance variable
        self.assertTrue(hasattr(TestThing, "not_a_class_prop"))
        self.assertEqual(instance.not_a_class_prop, 43)

        del TestThing.not_a_class_prop
        # descriptor itself is deleted
        self.assertFalse(hasattr(TestThing, "not_a_class_prop"))
        self.assertFalse(hasattr(instance, "not_a_class_prop"))
        with self.assertRaises(AttributeError):
            instance.not_a_class_prop

    def _generate_db_ops_tests(self) -> None:
        def test_prekill(thing: TestThing) -> None:
            self.assertEqual(thing.db_commit_number_prop, 0)
            thing.db_commit_number_prop = 100
            self.assertEqual(thing.db_commit_number_prop, 100)
            self.assertEqual(thing.db_engine.get_property("db_commit_number_prop"), 100)

            # test db persist property
            self.assertEqual(thing.db_persist_selector_prop, "a")
            thing.db_persist_selector_prop = "c"
            self.assertEqual(thing.db_persist_selector_prop, "c")
            self.assertEqual(thing.db_engine.get_property("db_persist_selector_prop"), "c")

            # test db init property
            self.assertEqual(thing.db_init_int_prop, TestThing.db_init_int_prop.default)
            thing.db_init_int_prop = 50
            self.assertEqual(thing.db_init_int_prop, 50)
            self.assertNotEqual(thing.db_engine.get_property("db_init_int_prop"), 50)
            self.assertEqual(
                thing.db_engine.get_property("db_init_int_prop"),
                TestThing.db_init_int_prop.default,
            )
            del thing

        def test_postkill(thing: TestThing) -> None:
            # deleted thing and reload from database
            self.assertEqual(thing.db_init_int_prop, TestThing.db_init_int_prop.default)
            self.assertEqual(thing.db_persist_selector_prop, "c")
            self.assertNotEqual(thing.db_commit_number_prop, 100)
            self.assertEqual(thing.db_commit_number_prop, TestThing.db_commit_number_prop.default)

        return test_prekill, test_postkill

    def test_06_sqlalchemy_db_operations(self):
        """Test SQLAlchemy database operations"""
        thing_id = "test-db-operations"
        file_path = f"{thing_id}.db"
        try:
            os.remove(file_path)
        except (OSError, FileNotFoundError):
            pass
        self.assertTrue(not os.path.exists(file_path))

        test_prekill, test_postkill = self._generate_db_ops_tests()

        thing = TestThing(id=thing_id, use_default_db=True, log_level=logging.WARN)
        test_prekill(thing)

        thing = TestThing(id=thing_id, use_default_db=True, log_level=logging.WARN)
        test_postkill(thing)

    def test_07_json_db_operations(self):
        with tempfile.NamedTemporaryFile(delete=False) as tf:
            filename = tf.name

        thing_id = "test-db-operations-json"
        test_prekill, test_postkill = self._generate_db_ops_tests()

        thing = TestThing(
            id=thing_id,
            use_json_file=True,
            json_filename=filename,
            log_level=logging.WARN,
        )
        test_prekill(thing)

        thing = TestThing(
            id=thing_id,
            use_json_file=True,
            json_filename=filename,
            log_level=logging.WARN,
        )
        test_postkill(thing)

        os.remove(filename)

    def test_08_db_config(self):
        """Test database configuration options"""
        thing = TestThing(id="test-sql-config", log_level=logging.WARN)

        # ----- SQL config tests -----
        sql_db_config = {
            "provider": "postgresql",
            "host": "localhost",
            "port": 5432,
            "database": "hololinked",
            "user": "hololinked",
            "password": "postgresnonadminpassword",
        }
        with open("test_sql_config.json", "w") as f:
            PythonBuiltinJSONSerializer.dump(sql_db_config, f)

        # correct config
        ThingDB(thing, config_file="test_sql_config.json")
        # foreign field
        sql_db_config_2 = copy.deepcopy(sql_db_config)
        sql_db_config_2["passworda"] = "postgresnonadminpassword"
        with open("test_sql_config.json", "w") as f:
            PythonBuiltinJSONSerializer.dump(sql_db_config_2, f)
        self.assertRaises(
            pydantic.ValidationError,
            ThingDB,
            thing,
            config_file="test_sql_config.json",
        )
        # missing field
        sql_db_config_3 = copy.deepcopy(sql_db_config)
        sql_db_config_3.pop("password")
        with open("test_sql_config.json", "w") as f:
            PythonBuiltinJSONSerializer.dump(sql_db_config_3, f)
        self.assertRaises(
            ValueError,
            ThingDB,
            thing,
            config_file="test_sql_config.json",
        )
        # URI instead of other fields
        sql_db_config = dict(
            provider="postgresql",
            uri="postgresql://hololinked:postgresnonadminpassword@localhost:5432/hololinked",
        )
        with open("test_sql_config.json", "w") as f:
            PythonBuiltinJSONSerializer.dump(sql_db_config, f)
        ThingDB(thing, config_file="test_sql_config.json")

        os.remove("test_sql_config.json")

        # ----- MongoDB config tests -----
        mongo_db_config = {
            "provider": "mongo",
            "host": "localhost",
            "port": 27017,
            "database": "hololinked",
            "user": "hololinked",
            "password": "mongononadminpassword",
            "authSource": "admin",
        }
        with open("test_mongo_config.json", "w") as f:
            PythonBuiltinJSONSerializer.dump(mongo_db_config, f)

        # correct config
        BaseDB.load_conf("test_mongo_config.json")
        # foreign field
        mongo_db_config_2 = copy.deepcopy(mongo_db_config)
        mongo_db_config_2["passworda"] = "mongononadminpassword"
        with open("test_mongo_config.json", "w") as f:
            PythonBuiltinJSONSerializer.dump(mongo_db_config_2, f)
        self.assertRaises(pydantic.ValidationError, BaseDB.load_conf, "test_mongo_config.json")
        # missing field
        mongo_db_config_3 = copy.deepcopy(mongo_db_config)
        mongo_db_config_3.pop("password")
        with open("test_mongo_config.json", "w") as f:
            PythonBuiltinJSONSerializer.dump(mongo_db_config_3, f)
        self.assertRaises(ValueError, BaseDB.load_conf, "test_mongo_config.json")
        # URI instead of other fields
        mongo_db_config = dict(
            provider="mongo",
            uri="mongodb://hololinked:mongononadminpassword@localhost:27017/hololinked?authSource=admin",
        )
        with open("test_mongo_config.json", "w") as f:
            PythonBuiltinJSONSerializer.dump(mongo_db_config, f)
        # correct config
        BaseDB.load_conf("test_mongo_config.json")

        os.remove("test_mongo_config.json")

        # ----- SQLite config tests -----

        sqlite_db_config = {
            "provider": "sqlite",
            "file": "test_sqlite.db",
        }
        with open("test_sqlite_config.json", "w") as f:
            PythonBuiltinJSONSerializer.dump(sqlite_db_config, f)

        # correct config
        ThingDB(thing, config_file="test_sqlite_config.json")

        os.remove("test_sqlite_config.json")


if __name__ == "__main__":
    unittest.main(testRunner=TestRunner())



"""
adopted from pyro - https://github.com/irmen/Pyro5 - see following license

MIT License

Copyright (c) Irmen de Jong

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import inspect
import array
import datetime
import io
import uuid
import decimal
import typing
import warnings
from enum import Enum
from collections import deque

# serializers:
import pickle
import json as pythonjson
from msgspec import json as msgspecjson, msgpack, Struct

# default dytypes:
try:
    import numpy
except ImportError:
    pass

from ..param.parameters import (
    TypeConstrainedList,
    TypeConstrainedDict,
    TypedKeyMappingsConstrainedDict,
    ClassSelector,
    String,
    Parameter,
)
from ..constants import JSONSerializable
from ..utils import MappableSingleton, format_exception_as_json, issubklass


class BaseSerializer(object):
    """
    Base class for (de)serializer implementations. All serializers must inherit this class
    and overload dumps() and loads() to be usable by the ZMQ message brokers. Any serializer
    that returns bytes when serialized and a python object on deserialization will be accepted.
    Serialization and deserialization errors will be passed as invalid message type
    (see ZMQ messaging contract) from server side and a exception will be raised on the client.
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = None

    def loads(self, data) -> typing.Any:
        "method called by ZMQ message brokers to deserialize data"
        raise NotImplementedError("implement loads()/deserialization in subclass")

    def dumps(self, data) -> bytes:
        "method called by ZMQ message brokers to serialize data"
        raise NotImplementedError("implement dumps()/serialization in subclass")

    def convert_to_bytes(self, data) -> bytes:
        if isinstance(data, bytes):
            return data
        if isinstance(data, bytearray):
            return bytes(data)
        if isinstance(data, memoryview):
            return data.tobytes()
        raise TypeError(
            "serializer convert_to_bytes accepts only bytes, bytearray or memoryview, not type {}".format(type(data))
        )

    @property
    def content_type(self) -> str:
        raise NotImplementedError("serializer must implement a content type")


dict_keys = type(dict().keys())


class JSONSerializer(BaseSerializer):
    "(de)serializer that wraps the msgspec JSON serialization protocol, default serializer for all clients."

    _type_replacements = {}

    def __init__(self) -> None:
        super().__init__()
        self.type = msgspecjson

    def loads(self, data: typing.Union[bytearray, memoryview, bytes]) -> JSONSerializable:
        "method called by ZMQ message brokers to deserialize data"
        return msgspecjson.decode(self.convert_to_bytes(data))

    def dumps(self, data) -> bytes:
        "method called by ZMQ message brokers to serialize data"
        return msgspecjson.encode(data, enc_hook=self.default)

    @classmethod
    def default(cls, obj) -> JSONSerializable:
        "method called if no serialization option was found."
        if hasattr(obj, "model_dump"):
            return obj.model_dump()
        if hasattr(obj, "json"):
            # alternative to type replacement
            return obj.json()
        if isinstance(obj, Struct):
            return obj
        if isinstance(obj, Enum):
            return obj.name
        if isinstance(obj, (set, dict_keys, deque, tuple)):
            # json module can't deal with sets so we make a tuple out of it
            return list(obj)
        if isinstance(obj, (TypeConstrainedDict, TypeConstrainedList, TypedKeyMappingsConstrainedDict)):
            return obj._inner  # copy has been implemented with same signature for both types
        if isinstance(obj, uuid.UUID):
            return str(obj)
        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()
        if isinstance(obj, decimal.Decimal):
            return str(obj)
        if isinstance(obj, Exception):
            return format_exception_as_json(obj)
        if isinstance(obj, array.array):
            if obj.typecode == "c":
                return obj.tostring()
            if obj.typecode == "u":
                return obj.tounicode()
            return obj.tolist()
        if "numpy" in globals() and isinstance(obj, numpy.ndarray):
            return obj.tolist()
        replacer = cls._type_replacements.get(type(obj), None)
        if replacer:
            return replacer(obj)
        raise TypeError("Given type cannot be converted to JSON : {}".format(type(obj)))

    @classmethod
    def register_type_replacement(cls, object_type, replacement_function) -> None:
        "register custom serialization function for a particular type"
        if object_type is type or not inspect.isclass(object_type):
            raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
        cls._type_replacements[object_type] = replacement_function

    @property
    def content_type(self) -> str:
        return "application/json"


class PythonBuiltinJSONSerializer(JSONSerializer):
    "(de)serializer that wraps the python builtin JSON serialization protocol."

    def __init__(self) -> None:
        super().__init__()
        self.type = pythonjson

    def loads(self, data: typing.Union[bytearray, memoryview, bytes]) -> typing.Any:
        "method called by ZMQ message brokers to deserialize data"
        return pythonjson.loads(self.convert_to_bytes(data))

    def dumps(self, data) -> bytes:
        "method called by ZMQ message brokers to serialize data"
        data = pythonjson.dumps(data, ensure_ascii=False, allow_nan=True, default=self.default)
        return data.encode("utf-8")

    @classmethod
    def dump(cls, data: typing.Dict[str, typing.Any], file_desc) -> None:
        "write JSON to file"
        pythonjson.dump(data, file_desc, ensure_ascii=False, allow_nan=True, default=cls.default)

    @classmethod
    def load(cls, file_desc) -> JSONSerializable:
        "load JSON from file"
        return pythonjson.load(file_desc)


class PickleSerializer(BaseSerializer):
    "(de)serializer that wraps the pickle serialization protocol, use with encryption for safety."

    def __init__(self) -> None:
        super().__init__()
        self.type = pickle

    def dumps(self, data) -> bytes:
        "method called by ZMQ message brokers to serialize data"
        return pickle.dumps(data)

    def loads(self, data) -> typing.Any:
        "method called by ZMQ message brokers to deserialize data"
        return pickle.loads(self.convert_to_bytes(data))

    @property
    def content_type(self) -> str:
        return "application/x-pickle"


class MsgpackSerializer(BaseSerializer):
    """
    (de)serializer that wraps the msgspec MessagePack serialization protocol, recommended serializer for ZMQ based
    high speed applications. Set an instance of this serializer to both ``Thing.zmq_serializer`` and
    ``hololinked.client.ObjectProxy``. Unfortunately, MessagePack is currently not supported for HTTP clients.
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = msgpack

    codes = dict(NDARRAY_EXT=1)

    def dumps(self, value) -> bytes:
        return msgpack.encode(value, enc_hook=self.default_encode)

    def loads(self, value) -> typing.Any:
        return msgpack.decode(self.convert_to_bytes(value), ext_hook=self.ext_decode)

    @classmethod
    def default_encode(cls, obj) -> typing.Any:
        if "numpy" in globals() and isinstance(obj, numpy.ndarray):
            buf = io.BytesIO()
            numpy.save(buf, obj, allow_pickle=False)  # use .npy. which stores dtype, shape, order, endianness
            return msgpack.Ext(MsgpackSerializer.codes["NDARRAY_EXT"], buf.getvalue())
        raise TypeError("Given type cannot be converted to MessagePack : {}".format(type(obj)))

    @classmethod
    def ext_decode(cls, code: int, obj: memoryview) -> typing.Any:
        if code == MsgpackSerializer.codes["NDARRAY_EXT"]:
            if "numpy" in globals():
                return numpy.load(io.BytesIO(obj), allow_pickle=False)
            else:
                raise ValueError("numpy is required to decode numpy array from MessagePack")
        return obj

    @property
    def content_type(self) -> str:
        return "application/msgpack"


class TextSerializer(BaseSerializer):
    """Converts string or string compatible types to bytes and vice versa"""

    def __init__(self) -> None:
        super().__init__()
        self.type = None

    def dumps(self, data) -> bytes:
        return str(data).encode("utf-8")

    def loads(self, data) -> typing.Any:
        return data.decode("utf-8")

    @property
    def content_type(self) -> str:
        return "text/plain"


try:
    import serpent

    class SerpentSerializer(BaseSerializer):
        """(de)serializer that wraps the serpent serialization protocol."""

        def __init__(self) -> None:
            super().__init__()
            self.type = serpent

        def dumps(self, data) -> bytes:
            "method called by ZMQ message brokers to serialize data"
            return serpent.dumps(data, module_in_classname=True)

        def loads(self, data) -> typing.Any:
            "method called by ZMQ message brokers to deserialize data"
            return serpent.loads(self.convert_to_bytes(data))

        @classmethod
        def register_type_replacement(cls, object_type, replacement_function) -> None:
            "register custom serialization function for a particular type"

            def custom_serializer(obj, serpent_serializer, outputstream, indentlevel):
                replaced = replacement_function(obj)
                if replaced is obj:
                    serpent_serializer.ser_default_class(replaced, outputstream, indentlevel)
                else:
                    serpent_serializer._serialize(replaced, outputstream, indentlevel)

            if object_type is type or not inspect.isclass(object_type):
                raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
            serpent.register_class(object_type, custom_serializer)

    # __all__.append(SerpentSerializer.__name__)
except ImportError:
    SerpentSerializer = None


class Serializers(metaclass=MappableSingleton):
    """
    A singleton class that holds all serializers and provides a registry for content types.
    All members are class attributes and settings are applied process-wide.

    - For property, the value is serialized using the serializer registered for the property.
    - For action, the return value is serialized using the serializer registered for the action.
    - For event, the payload is serialized using the serializer registered for the event.

    Registration of serializer is not mandatory for any property, action or event.
    The default serializer is `JSONSerializer`, which will be provided to any unregistered object.
    """

    json = ClassSelector(
        default=JSONSerializer(),
        class_=BaseSerializer,
        class_member=True,
        doc="The default serializer for all properties, actions and events",
    )  # type: BaseSerializer

    pickle = ClassSelector(
        default=PickleSerializer(),
        class_=BaseSerializer,
        class_member=True,
        doc="pickle serializer, unsafe without encryption but useful for faster & flexible serialization of python specific types",
    )  # type: BaseSerializer

    msgpack = ClassSelector(
        default=MsgpackSerializer(),
        class_=BaseSerializer,
        class_member=True,
        doc="MessagePack serializer, efficient binary format that is both fast & interoperable between languages ",
    )  # type: BaseSerializer

    text = ClassSelector(
        default=TextSerializer(),
        class_=BaseSerializer,
        class_member=True,
        doc="Text serializer, converts string or string compatible types to bytes and vice versa",
    )  # type: BaseSerializer

    default = ClassSelector(
        default=json.default, class_=BaseSerializer, class_member=True, doc="The default serialization to be used"
    )  # type: BaseSerializer

    default_content_type = String(
        fget=lambda self: self.default.content_type,
        class_member=True,
        doc="The default content type for the default serializer",
    )  # type: str

    content_types = Parameter(
        default={
            json.default.content_type: json.default,  # as in the default value of the descriptor
            pickle.default.content_type: pickle.default,
            msgpack.default.content_type: msgpack.default,
            text.default.content_type: text.default,
        },
        doc="A dictionary of content types and their serializers",
        readonly=True,
        class_member=True,
    )  # type: typing.Dict[str, BaseSerializer]

    allowed_content_types = Parameter(
        default=None,
        class_member=True,
        doc="A list of content types that are usually considered safe and will be supported by default without any configuration",
        readonly=True,
    )  # type: typing.List[str]

    object_content_type_map = Parameter(
        default=dict(),
        class_member=True,
        doc="A dictionary of content types for specific properties, actions and events",
        readonly=True,
    )  # type: typing.Dict[str, typing.Dict[str, str]]

    object_serializer_map = Parameter(
        default=dict(),
        class_member=True,
        doc="A dictionary of serializer for specific properties, actions and events",
        readonly=True,
    )  # type: typing.Dict[str, typing.Dict[str, BaseSerializer]]

    protocol_serializer_map = Parameter(
        default=dict(), class_member=True, doc="A dictionary of serializer for a specific protocol", readonly=True
    )  # type: typing.Dict[str, BaseSerializer]

    @classmethod
    def register(cls, serializer: BaseSerializer, name: str | None = None, override: bool = False) -> None:
        """
        Register a new serializer. It is recommended to implement a content type property/attribute for the serializer
        to facilitate automatic deserialization on client side, otherwise deserialization is not gauranteed by this implementation.

        Parameters
        ----------
        serializer: BaseSerializer
            the serializer to register
        name: str, optional
            the name of the serializer to be accessible under the object namespace. If not provided, the name of the
            serializer class is used.
        override: bool, optional
            whether to override the serializer if the content type is already registered, by default False & raises ValueError
            for duplicate content type.

        Raises
        ------
        ValueError
            if the serializer content type is already registered
        """
        try:
            if serializer.content_type in cls.content_types and not override:
                raise ValueError("content type already registered : {}".format(serializer.content_type))
            cls.content_types[serializer.content_type] = serializer
        except NotImplementedError:
            warnings.warn("serializer does not implement a content type", category=UserWarning)
        cls[name or serializer.__name__] = serializer

    @classmethod
    def for_object(cls, thing_id: str, thing_cls: str, objekt: str) -> BaseSerializer:
        """
        Retrieve a serializer for a given property, action or event

        Parameters
        ----------
        thing_id: str | typing.Any
            the id of the Thing or the Thing that owns the property, action or event
        thing_cls: str | typing.Any
            the class name of the Thing or the Thing that owns the property, action or event
        objekt: str | Property | Action | Event
            the name of the property, action or event

        Returns
        -------
        BaseSerializer | JSONSerializer
            the serializer for the property, action or event. If no serializer is found, the default JSONSerializer is
            returned.
        """
        if len(cls.object_serializer_map) == 0 and len(cls.object_content_type_map) == 0:
            return cls.default
        for thing in [thing_id, thing_cls]:  # first thing id, then thing cls
            if thing in cls.object_serializer_map:
                if objekt in cls.object_serializer_map[thing]:
                    return cls.object_serializer_map[thing][objekt]
                return cls.object_serializer_map[thing].get(thing, cls.default)
            if thing in cls.object_content_type_map:
                if objekt in cls.object_content_type_map[thing]:
                    return cls.content_types[cls.object_content_type_map[thing][objekt]]
                return cls.content_types[cls.object_content_type_map[thing].get(thing, cls.default_content_type)]
        return cls.default  # JSON is default serializer

    # @validate_call
    @classmethod
    def register_for_object(cls, objekt: typing.Any, serializer: BaseSerializer) -> None:
        """
        Register (an existing) serializer for a property, action or event. Other option is to register a content type,
        the effects are similar.

        Parameters
        ----------
        objekt: str | Property | Action | Event
            the property, action or event
        serializer: BaseSerializer
            the serializer to be used
        """
        if not isinstance(serializer, BaseSerializer):
            raise ValueError("serializer must be an instance of BaseSerializer, given : {}".format(type(serializer)))
        from ..core import Property, Action, Event, Thing

        if not isinstance(objekt, (Property, Action, Event)) and not issubklass(objekt, Thing):
            raise ValueError("object must be a Property, Action or Event, or Thing, got : {}".format(type(objekt)))
        if issubklass(objekt, Thing):
            owner = objekt.__name__
        elif not objekt.owner:
            raise ValueError("object owner cannot be determined : {}".format(objekt))
        else:
            owner = objekt.owner.__name__
        if owner not in cls.object_serializer_map:
            cls.object_serializer_map[owner] = dict()
        if issubklass(objekt, Thing):
            cls.object_serializer_map[owner][objekt.__name__] = serializer
        else:
            cls.object_serializer_map[owner][objekt.name] = serializer

    # @validate_call
    @classmethod
    def register_content_type_for_object(cls, objekt: typing.Any, content_type: str) -> None:
        """
        Register content type for a property, action, event, or a `Thing` class to use a specific serializer.

        Parameters
        ----------
        objekt: Property | Action | Event | Thing
            the property, action or event. string is not accepted - use `register_content_type_for_object_by_name()` instead.
        content_type: str
            the content type for the value of the objekt or the serializer to be used

        Raises
        ------
        ValueError
            if the object is not a Property, Action or Event
        """
        if content_type not in cls.content_types:
            raise ValueError("content type {} unsupported".format(content_type))
        from ..core import Property, Action, Event, Thing

        if not isinstance(objekt, (Property, Action, Event)) and not issubklass(objekt, Thing):
            raise ValueError("object must be a Property, Action or Event, got : {}".format(type(objekt)))
        if issubklass(objekt, Thing):
            owner = objekt.__name__
        elif not objekt.owner:
            raise ValueError("object owner cannot be determined, cannot register content type: {}".format(objekt))
        else:
            owner = objekt.owner.__name__
        if owner not in cls.object_content_type_map:
            cls.object_content_type_map[owner] = dict()
        if issubklass(objekt, Thing):
            cls.object_content_type_map[owner][objekt.__name__] = content_type
            # its a redundant key, TODO - may be there is a better way to structure this map
        else:
            cls.object_content_type_map[owner][objekt.name] = content_type

    # @validate_call
    @classmethod
    def register_content_type_for_object_per_thing_instance(
        cls, thing_id: str, objekt: str | typing.Any, content_type: str
    ) -> None:
        """
        Register an existing content type for a property, action or event to use a specific serializer. Other option is
        to register a serializer directly, the effects are similar.

        Parameters
        ----------
        thing_id: str
            the id of the Thing that owns the property, action or event
        objekt: str
            the name of the property, action or event
        content_type: str
            the content type to be used
        """
        if content_type not in cls.content_types:
            raise ValueError("content type {} unsupported".format(content_type))
        from ..core import Property, Action, Event

        if not isinstance(objekt, (Property, Action, Event, str)):
            raise ValueError("object must be a Property, Action or Event, got : {}".format(type(objekt)))
        if not isinstance(objekt, str):
            objekt = objekt.name
        if thing_id not in cls.object_content_type_map:
            cls.object_content_type_map[thing_id] = dict()
        cls.object_content_type_map[thing_id][objekt] = content_type

    @classmethod
    def register_content_type_for_thing_instance(cls, thing_id: str, content_type: str) -> None:
        """
        Register a content type for a specific Thing instance.

        Parameters
        ----------
        thing_id: str
            the id of the Thing
        content_type: str
            the content type to be used
        """
        if content_type not in cls.content_types:
            raise ValueError("content type {} unsupported".format(content_type))
        cls.object_content_type_map[thing_id][thing_id] = content_type
        # remember, its a redundant key, TODO

    @classmethod
    def register_for_object_per_thing_instance(cls, thing_id: str, objekt: str, serializer: BaseSerializer) -> None:
        """
        Register a serializer for a property, action or event for a specific Thing instance.

        Parameters
        ----------
        thing_id: str
            the id of the Thing that owns the property, action or event
        objekt: str
            the name of the property, action or event
        serializer: BaseSerializer
            the serializer to be used
        """
        if thing_id not in cls.object_serializer_map:
            cls.object_serializer_map[thing_id] = dict()
        cls.object_serializer_map[thing_id][objekt] = serializer

    @classmethod
    def register_for_thing_instance(cls, thing_id: str, serializer: BaseSerializer) -> None:
        """
        Register a serializer for a specific Thing instance.

        Parameters
        ----------
        thing_id: str
            the id of the Thing
        serializer: BaseSerializer
            the serializer to be used
        """
        if thing_id not in cls.object_serializer_map:
            cls.object_serializer_map[thing_id] = dict()
        cls.object_serializer_map[thing_id][thing_id] = serializer

    @classmethod
    def reset(cls) -> None:
        """
        Reset the serializer registry.
        """
        cls.object_content_type_map.clear()
        cls.object_serializer_map.clear()
        cls.protocol_serializer_map.clear()
        cls.default = cls.json

    @allowed_content_types.getter
    def get_allowed_content_types(cls) -> typing.List[str]:
        """
        Get a list of all allowed content types for serialization.
        """
        try:
            return cls._allowed_content_types
        except AttributeError:
            from ..config import global_config

            cls._allowed_content_types = list(cls.content_types.keys())
            cls._allowed_content_types.remove(cls.pickle.content_type)
            if global_config.ALLOW_PICKLE:
                cls._allowed_content_types.append(cls.pickle.content_type)
            return cls._allowed_content_types


__all__ = [
    JSONSerializer.__name__,
    PickleSerializer.__name__,
    MsgpackSerializer.__name__,
    TextSerializer.__name__,
    PythonBuiltinJSONSerializer.__name__,
    BaseSerializer.__name__,
    Serializers.__name__,
]



"""
Implementation of security schemes for a server.
"""

import base64


class Security:
    """Type definition for security schemes"""

    pass


try:
    import bcrypt

    class BcryptBasicSecurity(Security):
        """
        A username and password based security scheme using bcrypt.
        The password is stored as a hash and will be deleted from memory after initialization
        (most likely even after you keep a reference to the object).

        The request must supply an authorization header in of the following formats:

        - `Authorization: Basic base64(username:password)`
        - `Authorization: Basic (username:password)`

        The username and password are expected to be base64 encoded, by default.

        Set `expect_base64=False` if you want to use plain text credentials without base64 encoding.

        Use bcrypt when you are constrained in terms of memory. Use Argon2BasicSecurity
        when you can afford more memory (which is the recommended username-password implementation).

        Note: bcrypt is a cpu-hard hashing algorithm, which means it is resistant to brute-force attacks.
        """

        username: str
        _password_hash: bytes
        expect_base64: bool

        def __init__(self, username: str, password: str, expect_base64: bool = True) -> None:
            """
            Parameters
            ----------
            username: str
                The username to be used for authentication
            password: str
                The password to be used for authentication
            expect_base64: bool
                Whether to expect base64 encoded credentials in the authorization header. Default is True.
            """
            self.username = username
            self._password_hash = bcrypt.hashpw(password.encode("utf-8"), bcrypt.gensalt())
            del password  # Remove password from memory
            self.expect_base64 = expect_base64

        def validate(self, username: str, password: str) -> bool:
            """
            plain validate a username and password

            Returns
            -------
            bool
                True if the username and password are valid, False otherwise
            """
            if username != self.username:
                return False
            return bcrypt.checkpw(password.encode("utf-8"), self._password_hash)

        def validate_base64(self, b64_str: str) -> bool:
            """
            Validate a base64 encoded string containing username and password.
            Please strip the 'Basic ' prefix before passing.

            Returns
            -------
            bool
                True if the username and password are valid, False otherwise
            """
            if not self.expect_base64:
                raise ValueError("base64 encoded credentials not expected, please reconfigure if needed")
            try:
                decoded = base64.b64decode(b64_str).decode("utf-8")
            except (ValueError, TypeError):
                return False
            username, password = decoded.split(":", 1)
            return self.validate(username, password)

except ImportError:
    pass

try:
    import argon2

    class Argon2BasicSecurity(Security):
        """
        A username and password based security scheme using Argon2.
        The password is stored as a hash and will be deleted from memory after initialization
        (most likely even after you keep a reference to the object).

        The request must supply an authorization header in of the following formats:

        - `Authorization: Basic base64(username:password)`
        - `Authorization: Basic (username:password)`

        The username and password are expected to be base64 encoded, by default.

        Set `expect_base64=False` if you want to use plain text credentials without base64 encoding.

        Argon2 is the recommended password hashing security scheme.
        """

        username: str
        _password_hash: str
        expect_base64: bool

        def __init__(self, username: str, password: str, expect_base64: bool = True) -> None:
            self.ph = argon2.PasswordHasher()
            self.username = username
            self.expect_base64 = expect_base64
            self._password_hash = self.ph.hash(password)
            del password  # Remove password from memory

        def validate(self, username: str, password: str) -> bool:
            """
            plain validate a username and password

            Returns
            -------
            bool
                True if the username and password are valid, False otherwise
            """
            if username != self.username:
                return False
            try:
                return self.ph.verify(self._password_hash, password)
            except argon2.exceptions.VerifyMismatchError:
                return False

        def validate_base64(self, b64_str: str) -> bool:
            """
            Validate a base64 encoded string containing username and password.
            Please strip the 'Basic ' prefix before passing.

            Returns
            -------
            bool
                True if the username and password are valid, False otherwise
            """
            if not self.expect_base64:
                raise ValueError("base64 encoded credentials not expected, please reconfigure if needed")
            try:
                decoded = base64.b64decode(b64_str).decode("utf-8")
            except (ValueError, TypeError):
                return False
            username, password = decoded.split(":", 1)
            return self.validate(username, password)

except ImportError:
    pass
